import pandas as pd
import polars as pl


def match_df(
    description_file_path: str,
    df,
    hs_code_version: str,
    hs_code_column_name="hs_code",
    new_column_name="hs_code_description",
):
    is_pandas = False

    if isinstance(df, pd.DataFrame):
        df = pl.from_pandas(df)
        is_pandas = True

    # Read and filter the description file
    df_description = pl.read_parquet(description_file_path)
    df_description = df_description.filter(pl.col("version") == hs_code_version)

    # Perform join.
    df = df.join(
        df_description.select(
            pl.col("code").alias(hs_code_column_name),
            pl.col("description").alias(new_column_name),
        ),
        on=hs_code_column_name,
        how="left",
        coalesce=True,
    )

    if is_pandas:
        df = df.to_pandas()

    return df



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from dagster_pipes import (
    PipesContext,
    PipesDbfsContextLoader,
    PipesDbfsMessageWriter,
    PipesDefaultContextLoader,
    PipesDefaultMessageWriter,
    PipesEnvVarParamsLoader,
    open_dagster_pipes,
)
from pyspark.sql import SparkSession

from ascii_library.orchestration.pipes import Engine, ExecutionMode


class SparkScriptPipes(ABC):
    @abstractmethod
    def execute_business_logic(
        self,
        context,
        execution_mode: ExecutionMode,
        partition_key: Optional[str],
        spark: SparkSession,
        engine: Engine,
    ):
        raise NotImplementedError()

    def get_base_path_seeds(
        self,
        execution_mode: ExecutionMode,
    ) -> str:  # pyrefly: ignore
        if execution_mode == ExecutionMode.SmallDevSampleS3:
            return f"s3a://{self.sample_data}/ascii_seeds"
        elif execution_mode == ExecutionMode.SmallDevSampleLocal:
            return Path(self.reference_data_path, "ascii_seeds").resolve().as_uri()
        elif execution_mode == ExecutionMode.Full:
            return f"s3a://{self.bucket_seed_nodes}/ascii_seeds"

    def get_base_path_IO(
        self,
        execution_mode: ExecutionMode,
    ) -> str:  # pyrefly: ignore
        if execution_mode == ExecutionMode.SmallDevSampleS3:
            return f"s3a://{self.bucket_cc_dev_results}"
        elif execution_mode == ExecutionMode.SmallDevSampleLocal:
            return Path("z_state/ascii_dev_pipeline").resolve().as_uri()
        elif execution_mode == ExecutionMode.Full:
            return f"s3a://{self.bucket_cc_results}"

    def get_base_path_commoncrawl(
        self,
        execution_mode: ExecutionMode,
    ) -> str:  # pyrefly: ignore
        if execution_mode == ExecutionMode.SmallDevSampleS3:
            return f"s3a://{self.sample_data}/"
        elif execution_mode == ExecutionMode.SmallDevSampleLocal:
            return (
                Path("../../../../reference-data/pipeline_sample_data/commoncrawl")
                .resolve()
                .as_uri()
            )
        elif execution_mode == ExecutionMode.Full:
            return "s3a://commoncrawl/"

    def __init__(self):
        engine = Engine(os.environ.get("SPARK_PIPES_ENGINE", "pyspark"))
        loader, writer, envvar = self.get_loader_writer(engine)
        with open_dagster_pipes(
            context_loader=loader, message_writer=writer, params_loader=envvar
        ):
            context = PipesContext.get()
            execution_mode = ExecutionMode(context.get_extra("execution_mode"))  # type: ignore
            partition_key: Optional[str] = context.get_extra("partition_key")  # type: ignore

            context.log.info(f"Partition key: {partition_key}")
            context.log.info(f"Execution mode: {execution_mode}")

            self.bucket_seed_nodes = os.environ.get(
                "BUCKET_SEED_NODES", "ascii-supply-chain-research-input"
            )
            self.bucket_cc_results = os.environ.get(
                "BUCKET_CC_RESULTS", "ascii-supply-chain-research-results"
            )
            self.bucket_cc_dev_results = os.environ.get(
                "BUCKET_CC_DEV_RESULTS", "ascii-supply-chain-research-dev-results"
            )
            self.sample_data = os.environ.get(
                "SAMPLE_DATA_BUCKET", "ascii-supply-chain-research-sample-data"
            )
            self.reference_data_path = os.environ.get(
                "REFERENCE_DATA_PATH",
                "../../../../reference-data/pipeline_sample_data/commoncrawl",
            )

            if engine == Engine.Local:
                local_spark_config = context.get_extra("local_spark_config")
                spark_cfg = local_spark_config["spark_conf"]
                builder = SparkSession.builder.appName("ascii").master(  # type: ignore
                    spark_cfg["spark.master"]
                )
                for key, value in spark_cfg.items():
                    builder = builder.config(key, value)
                spark = builder.getOrCreate()
                # spark.sparkContext.setLogLevel("ERROR")
            elif engine in [Engine.Databricks, Engine.EMR]:
                # we are on a remote cluster
                # it is already initialized
                spark = SparkSession.builder.getOrCreate()  # type: ignore
            else:
                raise ValueError(f"Unsupported engine mode: {engine.value}")

            self.execute_business_logic(
                context, execution_mode, partition_key, spark, engine
            )

    def get_loader_writer(self, engine):
        if engine == Engine.Local:
            loader = PipesDefaultContextLoader()
            writer = PipesDefaultMessageWriter()
            envvar = PipesEnvVarParamsLoader()
        elif engine == Engine.Databricks:
            # databricks execution mode is on
            loader = PipesDbfsContextLoader()
            writer = PipesDbfsMessageWriter()
            envvar = PipesEnvVarParamsLoader()
        elif engine == Engine.EMR:
            import boto3
            from dagster_pipes import PipesS3ContextLoader, PipesS3MessageWriter

            # for this to work an instance profile with the right permissions
            # must already be attached to the cluster
            # this only works inside AWS
            s3_client = boto3.client(
                "s3",
                region_name="us-east-1",
            )
            loader = PipesS3ContextLoader(client=s3_client)
            writer = PipesS3MessageWriter(client=s3_client)
            envvar = PipesEnvVarParamsLoader()
        else:
            raise ValueError(f"Unsupported engine mode: {engine}")
        return loader, writer, envvar


# sample usage for child classes
# if __name__ == "__main__":
#    SparkScriptPipes()



from contextlib import contextmanager
from typing import Generator

import duckdb  # pyrefly: ignore
from duckdb import DuckDBPyConnection  # pyrefly: ignore
from pydantic import Field

from .duckdb_protocol import DuckDBConnectionProvider


class DuckDBFileResource(DuckDBConnectionProvider):
    """A Dagster resource that provides a connection to a local DuckDB file.

    Implements the DuckDBConnectionProvider protocol.
    """

    file_path: str = Field(
        description="Path to the DuckDB database file. Use ':memory:' for an in-memory database."
    )
    read_only: bool = Field(
        default=True, description="Whether to open the database in read-only mode."
    )

    @contextmanager
    def duckdb_connect(self) -> Generator[DuckDBPyConnection, None, None]:
        """Yields a connection to the specified DuckDB file."""
        if self.file_path == ":memory:":
            conn = duckdb.connect(database=self.file_path)
        else:
            conn = duckdb.connect(database=self.file_path, read_only=self.read_only)
        try:
            yield conn
        finally:
            conn.close()



import os

from dagster import DefaultSensorStatus, SensorDefinition
from dagster_slack import (
    make_slack_on_run_failure_sensor,
)

# see https://docs.dagster.io/_apidocs/libraries/dagster-slack


def make_slack_on_failure(base_url: str) -> SensorDefinition:
    return make_slack_on_run_failure_sensor(  # type: ignore
        channel="#ascii-pipeline-notifications",
        slack_token=os.environ.get("ALERT_NOTIFICATION_TOKEN", ""),
        webserver_base_url=base_url,
        default_status=DefaultSensorStatus.RUNNING,
    )



import os

from botocore.config import Config
from dagster import ConfigurableResource

from ascii_library.orchestration.pipes import Engine, ExecutionMode


class SparkPipesResource(ConfigurableResource):  # type: ignore
    r"""
    Generic configurable spark-pipes resource.

    Executes jobs in one of several modes: ``local`` for quick development,
    or on scalable cloud backends like ``databricks`` and ``emr``.
    Pipelines may optionally apply sampling to speed up end-to-end runs.

    **Databricks authentication (environment variables)**

    .. code-block:: text

       DATABRICKS_HOST
       DATABRICKS_CLIENT_ID
       DATABRICKS_CLIENT_SECRET

    **EMR credentials (environment variables)**

    .. code-block:: text

       ASCII_AWS_ACCESS_KEY_ID
       ASCII_AWS_SECRET_ACCESS_KEY

    :ivar engine: The default execution engine to use.
    :vartype engine: Engine
    :ivar execution_mode: The execution mode for the pipeline (e.g. ``debug``, ``prod``).
    :vartype execution_mode: ExecutionMode
    """

    engine: Engine
    execution_mode: ExecutionMode

    def get_spark_pipes_client(self, override_default_engine):
        aws_access_key_id = os.environ.get("ASCII_AWS_ACCESS_KEY_ID", "dummy")
        aws_secret_access_key = os.environ.get("ASCII_AWS_SECRET_ACCESS_KEY", "dummy")
        if override_default_engine is not None:
            engine_to_use = override_default_engine
        else:
            engine_to_use = self.engine
        if engine_to_use == Engine.Local:
            from dagster import PipesSubprocessClient

            return PipesSubprocessClient()
        elif engine_to_use == Engine.Databricks:
            import boto3
            from databricks.sdk import WorkspaceClient

            from ascii_library.orchestration.pipes.databricks import (
                PipesDatabricksEnhancedClient,
            )

            workspace_client = WorkspaceClient(
                host=os.environ.get("DATABRICKS_HOST", "dummy"),
                client_id=os.environ.get("DATABRICKS_CLIENT_ID", "dummy"),
                client_secret=os.environ.get("DATABRICKS_CLIENT_SECRET", "dummy"),
            )
            s3_client = boto3.client(
                "s3",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name="us-east-1",
            )
            tagging_client = boto3.client(
                "resourcegroupstaggingapi",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name="us-east-1",
            )
            return PipesDatabricksEnhancedClient(
                client=workspace_client,
                tagging_client=tagging_client,
                s3_client=s3_client,
            )
        elif engine_to_use == Engine.EMR:
            import boto3

            from ascii_library.orchestration.pipes.emr import PipesEmrEnhancedClient

            emrClient = boto3.client(
                "emr",
                region_name="us-east-1",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                config=Config(
                    retries={"max_attempts": 10, "mode": "adaptive"},
                ),
            )
            s3Client = boto3.client(
                "s3",
                region_name="us-east-1",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                config=Config(
                    retries={"max_attempts": 10, "mode": "adaptive"},
                ),
            )
            priceClient = boto3.client(
                "pricing",
                region_name="us-east-1",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                config=Config(
                    retries={"max_attempts": 10, "mode": "adaptive"},
                ),
            )
            return PipesEmrEnhancedClient(
                emr_client=emrClient,
                s3_client=s3Client,
                bucket="ascii-supply-chain-research-pipeline",
                price_client=priceClient,
            )
        else:
            raise ValueError(f"Unsupported engine mode: {engine_to_use}")



import os
from typing import Any, Dict, Optional

from dagster_pyspark import lazy_pyspark_resource


def get_pyspark_config(
    url: Optional[str] = None,
    AWS_ACCESS_KEY_ID: str = "ASCII_AWS_ACCESS_KEY_ID",
    AWS_SECRET_ACCESS_KEY: str = "ASCII_AWS_SECRET_ACCESS_KEY",
):
    return lazy_pyspark_resource.configured(
        dev_spark_config(url, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    )


def dev_spark_config(
    url: Optional[str] = None,
    AWS_ACCESS_KEY_ID: str = "ASCII_AWS_ACCESS_KEY_ID",
    AWS_SECRET_ACCESS_KEY: str = "ASCII_AWS_SECRET_ACCESS_KEY",
) -> Dict[str, Any]:
    try:
        key = os.environ[AWS_ACCESS_KEY_ID]
        secret = os.environ[AWS_SECRET_ACCESS_KEY]
    except KeyError:
        print(
            f"Please set {AWS_ACCESS_KEY_ID} and {AWS_SECRET_ACCESS_KEY} environment variables - using invalid default values instead"
        )
        key = "NO_KEY_PROVIDED"
        secret = "NO_SECRET_PROVIDED"
    results = {
        "spark_conf": {
            "spark.master": "local[*]",
            "spark.driver.memory": os.environ.get("SPARK_DRIVER_MEMORY", "16G"),
            "spark.sql.session.timeZone": "UTC",
            "spark.driver.extraJavaOptions": "-Duser.timezone=UTC",
            "spark.sql.adaptive.enabled": "true",
            "spark.sql.adaptive.skewedJoin.enabled": "true",
            "spark.sql.cbo.enabled": "true",
            "spark.sql.cbo.joinReorder.enabled": "true",
            "spark.sql.cbo.starSchemaDetection": "true",
            "spark.sql.autoBroadcastJoinThreshold": "500MB",
            "spark.driver.maxResultSize": "5G",
            "spark.sql.shuffle.partitions": "50",
            "spark.sql.statistics.histogram.enabled": "true",
            "spark.sql.execution.arrow.pyspark.enabled": "true",
            # TODO: consider fine tuning https://spark.apache.org/docs/latest/cloud-integration.html
            "spark.hadoop.fs.s3a.committer.name": "directory",
            # "spark.sql.sources.commitProtocolClass": "org.apache.spark.internal.io.cloud.PathOutputCommitProtocol"
            "spark.sql.parquet.output.committer.class": "org.apache.spark.internal.io.cloud.BindingParquetOutputCommitter",
            "spark.hadoop.fs.s3a.fast.upload": "true",
            "spark.hadoop.fs.s3a.path.style.access": "true",
            "spark.hadoop.fs.s3a.impl": "org.apache.hadoop.fs.s3a.S3AFileSystem",
            # "spark.hadoop.fs.s3a.endpoint": url,
            "spark.hadoop.fs.s3a.buffer.dir": os.environ.get(
                "SPARK_DIR", "/data/raid5/data/sparktmp"
            ),
            "spark.hadoop.fs.s3a.fast.upload.buffer": "bytebuffer",
            "spark.hadoop.fs.s3a.fast.upload.active.blocks": "4",
            # "spark.databricks.delta.schema.autoMerge.enabled": "true",
            # "spark.sql.parquet.mergeSchema": "true",
            "spark.sql.parquet.compression.codec": "gzip",
            "spark.hadoop.fs.s3a.access.key": key,
            "spark.hadoop.fs.s3a.secret.key": secret,
            "spark.sql.extensions": "io.delta.sql.DeltaSparkSessionExtension",
            "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            # org.apache.hadoop:hadoop-aws:3.4.2
            "spark.jars.packages": "io.delta:delta-spark_2.12:3.3.2,org.postgresql:postgresql:42.7.7,org.apache.spark:spark-hadoop-cloud_2.12:3.5.6,com.johnsnowlabs.nlp:spark-nlp_2.12:6.1.3",
            "spark.databricks.delta.schema.autoMerge.enabled": "True",
            "spark.databricks.delta.schema.autoMerge.enabledOnWrite": "True",
            "spark.local.dir": os.environ.get("SPARK_DIR", "/data/raid5/data/sparktmp"),
        }
    }
    if url:
        results["spark_conf"]["spark.hadoop.fs.s3a.endpoint"] = url
    return results



import os
import sys
import time
from typing import Any, Dict, List, Mapping, Optional

from dagster import (
    OpExecutionContext,
    PipesContextInjector,
    PipesMessageReader,
    ResourceParam,
    get_dagster_logger,
    open_pipes_session,
)
from dagster._core.pipes.client import PipesClientCompletedInvocation  # type: ignore
from dagster_databricks import (
    PipesDbfsContextInjector,
    PipesDbfsLogReader,
    PipesDbfsMessageReader,
)
from dagster_pipes import PipesExtras
from dagster_shared.check.functions import bool_param, opt_inst_param
from databricks.sdk import WorkspaceClient
from databricks.sdk.service import jobs
from mypy_boto3_resourcegroupstaggingapi import ResourceGroupsTaggingAPIClient
from pydantic import Field

from ascii_library.orchestration.pipes.cloud_client import _PipesBaseCloudClient
from ascii_library.orchestration.pipes.exceptions import CustomPipesException


class _PipesDatabricksClient(_PipesBaseCloudClient):
    """Pipes client for Databricks.

    :param client: A Databricks ``WorkspaceClient`` object.
    :type client: WorkspaceClient
    :param tagging_client: A Boto3 client for resource tagging.
    :type tagging_client: ResourceGroupsTaggingAPIClient
    :param context_injector: A context injector to use to inject context into the
        Databricks job. Defaults to a configured ``PipesDbfsContextInjector``.
    :type context_injector: Optional[PipesContextInjector]
    :param message_reader: A message reader to use to read messages from the
        Databricks job. Defaults to a configured ``PipesDbfsMessageReader``.
    :type message_reader: Optional[PipesMessageReader]
    :param forward_termination: If True, the Databricks job will be canceled if the
        orchestration process is interrupted. Defaults to True.
    :type forward_termination: bool
    :param kwargs: Additional keyword arguments.
    :type kwargs: Any
    """

    env: Optional[Mapping[str, str]] = Field(
        default=None,
        description="An optional dict of environment variables to pass to the subprocess.",
    )

    def __init__(
        self,
        client: WorkspaceClient,
        tagging_client: ResourceGroupsTaggingAPIClient,
        context_injector: Optional[PipesContextInjector] = None,
        message_reader: Optional[PipesMessageReader] = None,
        forward_termination: bool = True,
        **kwargs,
    ):
        super().__init__(
            main_client=client,
            context_injector=context_injector,
            message_reader=message_reader,
            tagging_client=tagging_client,
        )
        self._s3_client = kwargs.get("s3_client")
        self._tagging_client = tagging_client
        self.client = client
        self.context_injector = opt_inst_param(
            context_injector,
            "context_injector",
            PipesContextInjector,
        ) or PipesDbfsContextInjector(client=self.client)
        self.message_reader = opt_inst_param(
            message_reader,
            "message_reader",
            PipesMessageReader,
        )
        self.forward_termination = bool_param(
            forward_termination, "forward_termination"
        )

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return False

    def get_default_message_reader(
        self, task: jobs.SubmitTask
    ) -> "PipesDbfsMessageReader":
        # include log readers if the user is writing their logs to DBFS
        if (
            task.as_dict()
            .get("new_cluster", {})
            .get("cluster_log_conf", {})
            .get("dbfs", None)
        ):
            log_readers = [
                PipesDbfsLogReader(
                    client=self.client,
                    remote_log_name="stdout",
                    target_stream=sys.stdout,
                ),
                PipesDbfsLogReader(
                    client=self.client,
                    remote_log_name="stderr",
                    target_stream=sys.stderr,
                ),
            ]
        else:
            log_readers = None
        return PipesDbfsMessageReader(
            client=self.client,
            log_readers=log_readers,
        )

    def _prepare_environment(
        self,
        local_file_path: str,
        dbfs_path: str,
        libraries_to_build_and_upload: Optional[List[str]],
    ) -> None:
        """Prepare the environment by uploading files and ensuring libraries are available."""
        self._upload_file_to_cloud(
            local_file_path=local_file_path, cloud_path=dbfs_path
        )
        self.filesystem = "s3"
        self._ensure_library_on_cloud(
            libraries_to_build_and_upload=libraries_to_build_and_upload
        )

    def _process_submit_args(
        self, submit_args: Optional[Mapping[str, Any]]
    ) -> Dict[str, Any]:
        """Process submit_args to ensure they are of the allowed types."""
        if not submit_args:
            return {}

        allowed_types = (
            list,
            jobs.JobEmailNotifications,
            jobs.GitSource,
            jobs.JobsHealthRules,
            jobs.JobNotificationSettings,
            jobs.QueueSettings,
            int,
            jobs.WebhookNotifications,
        )
        submit_kwargs = {}
        for key, value in submit_args.items():
            if isinstance(value, allowed_types):
                submit_kwargs[key] = value
            else:
                raise TypeError(f"Unexpected type for submit_arg {key}: {type(value)}")
        return submit_kwargs

    def run(  # type: ignore
        self,
        env: Optional[Mapping[str, str]],
        context: OpExecutionContext,
        extras: Optional[PipesExtras],
        task: jobs.SubmitTask,
        submit_args: Optional[Mapping[str, str]],
        local_file_path: str,
        dbfs_path: str,
        libraries_to_build_and_upload: Optional[List[str]] = None,
    ) -> PipesClientCompletedInvocation:
        """Synchronously execute a Databricks job with the pipes protocol.

        :param env: An optional dict of environment variables to pass.
        :type env: Optional[Mapping[str, str]]
        :param context: The context from the executing op or asset.
        :type context: OpExecutionContext
        :param extras: An optional dict of extra parameters to pass to the subprocess.
        :type extras: Optional[PipesExtras]
        :param task: Specification of the Databricks task to run.
        :type task: databricks.sdk.service.jobs.SubmitTask
        :param submit_args: Additional keyword arguments that will be forwarded
                            as-is to ``WorkspaceClient.jobs.submit``.
        :type submit_args: Optional[Mapping[str, str]]
        :param local_file_path: The local path to the script to be executed.
        :type local_file_path: str
        :param dbfs_path: The corresponding path on DBFS where the script will be uploaded.
        :type dbfs_path: str
        :param libraries_to_build_and_upload: A list of local Python packages to build
                                            and upload as libraries.
        :type libraries_to_build_and_upload: Optional[List[str]]

        :return: Wrapper containing results reported by the external process.
        :rtype: PipesClientCompletedInvocation
        """
        self._prepare_environment(
            local_file_path, dbfs_path, libraries_to_build_and_upload
        )
        submit_kwargs = self._process_submit_args(submit_args)

        message_reader = self.message_reader or self.get_default_message_reader(task)
        with open_pipes_session(
            context=context,
            extras=extras,
            context_injector=self.context_injector,
            message_reader=message_reader,
        ) as pipes_session:
            submit_task_dict = task.as_dict()
            ascii_wandb_value = {"ASCII_WANDB": os.environ.get("ASCII_WANDB", "")}
            if not ascii_wandb_value:
                get_dagster_logger().warning(
                    "Environment variable 'ASCII_WANDB' is not set; defaulting to empty value."
                )
            submit_task_dict["new_cluster"]["spark_env_vars"] = {
                **submit_task_dict["new_cluster"].get("spark_env_vars", {}),
                **(env or {}),
                **pipes_session.get_bootstrap_env_vars(),
                **ascii_wandb_value,
            }
            task = jobs.SubmitTask.from_dict(submit_task_dict)
            submission = self.client.jobs.submit(
                run_name=extras.get("job_name"),  # type: ignore
                tasks=[task],
                **(submit_kwargs or {}),
            )
            run_id = submission.run_id
            context.log.info(
                f"Databricks url: {self.client.jobs.get_run(run_id).run_page_url}"
            )
            try:
                self._poll_till_success(
                    run_id=run_id, extras=extras, tagging_client=self._tagging_client
                )
            except CustomPipesException:
                if self.forward_termination:
                    context.log.info(
                        "[pipes] execution interrupted, canceling Databricks job."
                    )
                    self.client.jobs.cancel_run(run_id)
                    self._poll_til_terminating(str(run_id))
                raise
        return PipesClientCompletedInvocation(pipes_session)

    def _poll_til_terminating(self, run_id: str) -> None:
        # Wait to see the job enters a state that indicates the underlying task is no longer executing
        # TERMINATING: "The task of this run has completed, and the cluster and execution context are being cleaned up."
        run_id_int = int(run_id)
        while True:
            run = self.client.jobs.get_run(run_id_int)
            if (
                run
                and run.state
                and run.state.life_cycle_state
                in (
                    jobs.RunLifeCycleState.TERMINATING,
                    jobs.RunLifeCycleState.TERMINATED,
                    jobs.RunLifeCycleState.SKIPPED,
                    jobs.RunLifeCycleState.INTERNAL_ERROR,
                )
            ):
                return

            time.sleep(self.poll_interval_seconds)


PipesDatabricksEnhancedClient = ResourceParam[_PipesDatabricksClient]



import base64
import time
from typing import List, Optional, Union

from botocore.client import BaseClient
from botocore.exceptions import (
    ClientError,
    ConnectionError,
    ConnectTimeoutError,
    NoCredentialsError,
    ReadTimeoutError,
    ResponseStreamingError,
)
from dagster import (
    PipesClient,
    PipesContextInjector,
    PipesMessageReader,
    file_relative_path,
    get_dagster_logger,
)
from dagster_shared.check.functions import numeric_param, opt_inst_param
from databricks.sdk import WorkspaceClient
from databricks.sdk.core import DatabricksError
from databricks.sdk.service import jobs
from tenacity import (
    RetryCallState,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
)

from ascii_library.orchestration.pipes.exceptions import CustomPipesException
from ascii_library.orchestration.resources.emr_constants import pipeline_bucket

from .utils import library_to_cloud_paths, package_library


def after_retry(retry_state: RetryCallState):
    """Function to log after each retry attempt."""
    if retry_state.next_action is not None:
        sleep_time = retry_state.next_action.sleep
    else:
        sleep_time = 0.0
    get_dagster_logger().debug(
        (
            f"Retry attempt: {retry_state.attempt_number}. Waiting {sleep_time} seconds before next try."
        )
    )


class _PipesBaseCloudClient(PipesClient):
    """
    Base class for Pipes clients, containing common methods and attributes.
    """

    def __init__(
        self,
        main_client: Union[BaseClient, WorkspaceClient],
        context_injector: Optional[PipesContextInjector] = None,
        message_reader: Optional[PipesMessageReader] = None,
        poll_interval_seconds: float = 5,
        **kwargs,
    ):
        self.poll_interval_seconds = numeric_param(
            poll_interval_seconds, "poll_interval_seconds"
        )
        self.message_reader = opt_inst_param(
            message_reader,
            "message_reader",
            PipesMessageReader,
        )
        self.main_client = main_client
        self.last_observed_state = None
        if (
            isinstance(main_client, BaseClient)
            and main_client.meta.service_model.service_name == "emr"
        ):
            self._s3_client = kwargs.get("s3_client")
            self.filesystem = ""
        elif isinstance(main_client, WorkspaceClient):
            self.filesystem = "dbfs"
            self._tagging_client = kwargs.get("tagging_client")

    @retry(
        stop=stop_after_delay(20) | stop_after_attempt(10),
        wait=wait_exponential(multiplier=1, max=60),
        after=after_retry,
        retry=retry_if_exception_type(
            (
                ConnectTimeoutError,
                ReadTimeoutError,
                ResponseStreamingError,
                ConnectionError,
            )
        ),
    )
    def _retrieve_state_emr(self, cluster_id):
        return self.main_client.describe_cluster(ClusterId=cluster_id)  # type: ignore

    def _handle_emr_polling(self, cluster_id):
        description = self._retrieve_state_emr(cluster_id)
        state = description["Cluster"]["Status"]["State"]
        dns = description["Cluster"].get("MasterPublicDnsName")  # Correct this part
        if state != self.last_observed_state:
            if dns:
                get_dagster_logger().debug(f"dns: {dns}")
            get_dagster_logger().info(
                f"[pipes] EMR cluster id {cluster_id} observed state transition to {state}"
            )
            self.last_observed_state = state
        if state in ["TERMINATED", "TERMINATING", "TERMINATED_WITH_ERRORS"]:
            return self._handle_terminated_state_emr(
                job_flow=cluster_id, description=description, state=state
            )
        else:
            return True

    @retry(
        stop=stop_after_delay(20) | stop_after_attempt(10),
        wait=wait_exponential(multiplier=1, max=60),
        after=after_retry,
        retry=retry_if_exception_type((DatabricksError)),
    )
    def _retrieve_state_dbr(self, run_id) -> jobs.Run:
        return self.main_client.jobs.get_run(run_id)  # pyrefly: ignore

    def _handle_dbr_polling(self, run_id):
        run = self._retrieve_state_dbr(run_id)
        state = run.state
        assert isinstance(state, jobs.RunState)
        if state.life_cycle_state != self.last_observed_state:
            get_dagster_logger().debug(
                f"[pipes] Databricks run {run_id} observed state transition to {state.life_cycle_state}"
            )
            self.last_observed_state = state.life_cycle_state  # pyrefly: ignore
        if state.life_cycle_state in (
            jobs.RunLifeCycleState.TERMINATED,
            jobs.RunLifeCycleState.SKIPPED,
            jobs.RunLifeCycleState.INTERNAL_ERROR,
            jobs.RunLifeCycleState.TERMINATING,
        ):
            get_dagster_logger().debug(f"Handling terminated state for run: {run_id}")
            return self._handle_terminated_state_dbr(run=run)
        else:
            return True

    def _poll_till_success(self, **kwargs):
        self._tagging_client = kwargs.get("tagging_client")
        cont = True
        if kwargs.get("extras") is not None:
            self.engine = kwargs.get("extras")["engine"]  # type: ignore
            self.executionMode = kwargs.get("extras")["execution_mode"]  # type: ignore
        while cont:
            if isinstance(self.main_client, BaseClient):
                cont = self._handle_emr_polling(kwargs.get("cluster_id"))
            elif isinstance(self.main_client, WorkspaceClient):
                cont = self._handle_dbr_polling(run_id=kwargs.get("run_id"))
            else:
                raise TypeError(
                    "main_client must be either EMRClient or WorkspaceClient"
                )
            time.sleep(self.poll_interval_seconds)

    def _handle_terminated_state_emr(self, job_flow, description, state):
        reason = description["Cluster"]["Status"].get("StateChangeReason", {}) or {}
        code = (reason.get("Code") or "").upper()
        message = reason.get("Message") or ""
        if (
            "error" in message.lower()
            or "failed" in message.lower()
            or state == "TERMINATED_WITH_ERRORS"
        ):
            raise CustomPipesException(
                message=f"Error running EMR job flow: {job_flow}"
            )
        elif (
            state == "TERMINATING" or state == "TERMINATED"
        ) and code == "ALL_STEPS_COMPLETED":
            return False
        else:
            return True

    def _handle_terminated_state_emr(self, job_flow, description, state):
        status = description["Cluster"]["Status"]
        reason = status.get("StateChangeReason", {}) or {}
        code = (reason.get("Code") or "").upper()
        message = reason.get("Message") or ""

        if state == "TERMINATED_WITH_ERRORS":
            raise CustomPipesException(
                message=f"EMR job {job_flow} failed: [{code}] {message}"
            )

        if state in ("TERMINATED", "TERMINATING"):
            if code == "ALL_STEPS_COMPLETED":
                return False  # success -> stop polling
            log_uri = description["Cluster"].get("LogUri") or ""
            hint = f" (logs: {log_uri})" if log_uri else ""
            raise CustomPipesException(
                message=f"EMR job {job_flow} ended in {state} unexpectedly: [{code}] {message}{hint}"
            )

        return True

    def _handle_terminated_state_dbr(self, run):
        client = self._tagging_client
        get_dagster_logger().debug(f"run: {run}")
        try:
            resources = client.get_resources(
                TagFilters=[
                    {
                        "Key": "JobId",
                        "Values": [
                            str(run.job_id),
                        ],
                    },
                ]
            )
            resource_arns = [
                item["ResourceARN"]  # type: ignore[index]
                for item in resources["ResourceTagMappingList"]
            ]
            if len(resource_arns) > 0:
                for arn in resource_arns:
                    client.tag_resources(
                        ResourceARNList=[arn],
                        Tags={
                            "jobId": str(run.job_id),
                            "engine": self.engine,
                            "executionMode": self.executionMode,
                        },
                    )
        except Exception as e:
            get_dagster_logger().debug(e)
            raise
        if run.state.result_state == jobs.RunResultState.SUCCESS:
            return False
        else:
            state_message = run.state.state_message or "Unknown reason"
            raise CustomPipesException(
                message=f"Error running Databricks job: {state_message}"
            )

    def _ensure_library_on_cloud(
        self,
        libraries_to_build_and_upload: Optional[List[str]],
        **kwargs,
    ):
        """
        Ensure that the specified library is available on S3.

        Args:
            dbfs_path (str): The S3 path where the library should reside.
            local_library_path (str): The local file system path to the library.
        """
        bucket = kwargs.get("bucket", pipeline_bucket)
        if libraries_to_build_and_upload is not None:
            for library in libraries_to_build_and_upload:
                path = library_to_cloud_paths(
                    lib_name=library, filesystem=self.filesystem
                )
                to_upload = package_library(
                    file_relative_path(__file__, f"../../../../{library}")
                )[0]
                self._upload_file_to_cloud(
                    local_file_path=to_upload, cloud_path=path, bucket=bucket
                )

    def _upload_file_to_cloud(self, local_file_path: str, cloud_path: str, **kwargs):
        try:
            if isinstance(self.main_client, BaseClient):
                self._upload_file_to_s3(local_file_path, cloud_path, **kwargs)
            elif (
                isinstance(self.main_client, WorkspaceClient)
                and "py" == local_file_path.split(".")[-1]
            ):
                self._upload_file_to_dbfs(local_file_path, cloud_path)
            elif (
                isinstance(self.main_client, WorkspaceClient)
                and "whl" == local_file_path.split(".")[-1]
            ):
                self._upload_file_to_s3(local_file_path, cloud_path, **kwargs)
            else:
                raise TypeError(
                    "main_client must be either EMRClient or WorkspaceClient"
                )
        except Exception as e:
            self.handle_exep(e)

    def handle_exep(self, e):
        if isinstance(e, FileNotFoundError):
            get_dagster_logger().error("The file was not found")
            raise e
        elif isinstance(e, NoCredentialsError):
            get_dagster_logger().error("Credentials not available")
            raise e
        elif isinstance(e, ClientError):
            get_dagster_logger().error("Client error while uploading")
            raise e

    def _upload_file_to_s3(self, local_file_path: str, cloud_path: str, **kwargs):
        bucket = kwargs.get("bucket")
        get_dagster_logger().debug(f"uploading: {cloud_path} into bucket: {bucket}")
        try:
            if self._s3_client is not None:
                self._s3_client.upload_file(local_file_path, bucket, cloud_path)
            else:  # noqa: E722
                get_dagster_logger().debug("fail to upload to S3")
        except Exception as e:
            get_dagster_logger().error(f"error: {e.with_traceback}")
            raise

    def _upload_file_to_dbfs(self, local_file_path: str, dbfs_path: str):
        get_dagster_logger().debug(
            f"uploading: {local_file_path} to DBFS at: {dbfs_path}"
        )
        try:
            assert isinstance(self.main_client, WorkspaceClient)
            with open(local_file_path, "rb") as file:
                encoded_string = base64.b64encode(file.read()).decode("utf-8")
            self.main_client.dbfs.put(
                path=dbfs_path, contents=encoded_string, overwrite=True
            )
        except Exception as e:
            get_dagster_logger().error(e.with_traceback)
            get_dagster_logger().error(e)
            raise



import html as std_html
import re

import bs4
from bs4 import BeautifulSoup
from loguru import logger

# Configure Loguru Logger (optional)
# logger.remove()  # Remove default logger to prevent duplicate logs
# logger.add(
#     "text_extractor.log",
#     rotation="10 MB",
#     level="ERROR",
#     format="{time} {level} {message}",
# )


class TextExtractor:
    """
    A class that traverses an HTML parse tree to extract text content, handling various HTML elements.
    """

    BLOCK_ELEMENTS = {
        "ADDRESS",
        "ARTICLE",
        "ASIDE",
        "BLOCKQUOTE",
        "BODY",
        "BR",
        "BUTTON",
        "CANVAS",
        "CAPTION",
        "COL",
        "COLGROUP",
        "DD",
        "DIV",
        "DL",
        "DT",
        "EMBED",
        "FIELDSET",
        "FIGCAPTION",
        "FIGURE",
        "FOOTER",
        "FORM",
        "H1",
        "H2",
        "H3",
        "H4",
        "H5",
        "H6",
        "HEADER",
        "HGROUP",
        "HR",
        "LI",
        "MAP",
        "NOSCRIPT",
        "OBJECT",
        "OL",
        "OUTPUT",
        "P",
        "PRE",
        "PROGRESS",
        "SECTION",
        "TABLE",
        "TBODY",
        "TEXTAREA",
        "TFOOT",
        "TH",
        "THEAD",
        "TR",
        "UL",
        "VIDEO",
    }
    INLINE_SPACING_ELEMENTS = {
        "ADDRESS",
        "CITE",
        "DETAILS",
        "DATALIST",
        "IFRAME",
        "IMG",
        "INPUT",
        "LABEL",
        "LEGEND",
        "OPTGROUP",
        "Q",
        "SELECT",
        "SUMMARY",
        "TBODY",
        "TD",
        "TIME",
        "SPAN",
        "A",
    }
    HEADING_TAGS = {"H1", "H2", "H3", "H4", "H5", "H6"}
    UNWANTED_ELEMENTS = {"SCRIPT", "STYLE", "HEAD", "META", "NOSCRIPT"}

    def __init__(self):
        self.text_extract = []  # List to accumulate text segments
        self.extracted_text = None  # Will be set to final extracted text or None
        self.in_pre = False  # Indicates if we're inside a <pre> tag

    def handle_node(self, node):  # noqa: C901
        """
        Recursively handle each node in the HTML tree.
        """
        if isinstance(node, BeautifulSoup):
            # Root of the soup
            for child in node.contents:
                self.handle_node(child)
        elif isinstance(node, bs4.element.Tag):
            name = self.get_node_tag_name(node)
            if name in self.UNWANTED_ELEMENTS:
                return  # Skip this node and its children
            self.handle_tag_open(name)

            # Process children
            for child in node.contents:
                self.handle_node(child)

            self.handle_tag_close(name)
        elif isinstance(node, bs4.element.NavigableString):
            if not isinstance(node, bs4.element.Comment):
                self.handle_text_node(node)

    def get_node_tag_name(self, node):
        """
        Safely get the uppercased tag name of an element node.
        """
        try:
            name = node.name
            return name.upper() if name else None
        except Exception as e:
            logger.error(f"Error processing node tag: {e}")
            return None

    def handle_tag_open(self, name):
        """
        Handle the opening of an HTML tag.
        """
        if name == "PRE":
            self.in_pre = True

        if name in self.BLOCK_ELEMENTS:
            self.append_paragraph_separator()
        elif name in self.INLINE_SPACING_ELEMENTS:
            self.append_space()

    def handle_tag_close(self, name):
        """
        Handle the closing of an HTML tag.
        """
        if name == "PRE":
            self.in_pre = False

        if name in self.BLOCK_ELEMENTS:
            self.append_paragraph_separator()
        elif name in self.INLINE_SPACING_ELEMENTS:
            self.append_space()

        if name in self.HEADING_TAGS:
            self.append_paragraph_separator()

    def handle_text_node(self, text_node):
        """
        Handle the text content of an HTML node.
        """
        try:
            txt = self.decode_char_ent(str(text_node))
            if self.in_pre:
                t = txt  # Preserve whitespace in <pre> tags
            else:
                t = " ".join(txt.split())

            if not t.strip():
                return  # Skip empty strings

            self.text_extract.append(t + " ")  # Append a space after the text
        except Exception as e:
            logger.error(f"Error handling text node: {e}")

    def append_paragraph_separator(self):
        """
        Append a paragraph separator (newline) to the text extract.
        """
        if self.text_extract and not self.text_extract[-1].endswith("\n"):
            self.text_extract.append("\n")
        else:
            self.text_extract.append("\n")

    def append_space(self):
        """
        Append a space to the text extract if appropriate.
        """
        if self.text_extract:
            if not self.text_extract[-1].endswith((" ", "\n")):
                self.text_extract.append(" ")
        else:
            self.text_extract.append(" ")

    def decode_char_ent(self, text):
        """
        Decode HTML character entities.
        """
        return std_html.unescape(text)

    def get_extracted_text(self):
        """
        After processing, get the final extracted text.
        """
        if self.text_extract:
            # Join the text segments and perform final cleanup
            final_text = "".join(self.text_extract)
            # Remove text enclosed in curly braces {}
            final_text = re.sub(r"\{.*?\}", "", final_text, flags=re.DOTALL)
            # Replace multiple newlines with a single newline
            final_text = re.sub(r"\n+", "\n", final_text)
            # Replace multiple spaces with a single space
            final_text = re.sub(r"[ ]{2,}", " ", final_text)
            self.extracted_text = final_text.strip()  # pyrefly: ignore
        else:
            self.extracted_text = None  # No text extracted



import os
from contextlib import contextmanager
from typing import Generator, List, Literal, Tuple, Union

import dagster as dg
import duckdb  # pyrefly: ignore
from duckdb import DuckDBPyConnection  # pyrefly: ignore
from pydantic import Field
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Connection, Engine

from .duckdb_protocol import DuckDBConnectionProvider


class PostgresConfig(dg.Config):
    """Configuration for a Postgres metadata backend."""

    type: Literal["postgres"] = "postgres"
    host: str = Field(default=os.getenv("DUCKLAKE_PG_HOST", "localhost"))
    port: int = Field(default=5432)
    database: str
    user: str
    password: str

    def get_ducklake_sql_parts(self, alias: str) -> Tuple[str, str]:
        """Returns the SQL for the credential secret and the main METADATA parameter."""
        secret_name = f"secret_catalog_{alias}"
        secret_sql = f"""
            CREATE OR REPLACE SECRET {secret_name} (
                TYPE postgres, HOST '{self.host}', PORT {self.port},
                DATABASE '{self.database}', USER '{self.user}',
                PASSWORD '{self.password}'
            );
        """
        metadata_params_sql = (
            f"METADATA_PATH '', "
            f"METADATA_PARAMETERS MAP {{'TYPE': 'postgres', 'SECRET': '{secret_name}'}}"
        )
        return secret_sql, metadata_params_sql


class SqliteConfig(dg.Config):
    """Configuration for a local SQLite file metadata backend."""

    type: Literal["sqlite"] = "sqlite"
    path: str = Field(description="Path to the SQLite database file.")

    def get_ducklake_sql_parts(self, alias: str) -> Tuple[str, str]:
        """For file-based backends, no credential secret is needed."""
        return "", f"METADATA_PATH '{self.path}'"


class DuckDBConfig(dg.Config):
    """Configuration for a local DuckDB file metadata backend."""

    type: Literal["duckdb"] = "duckdb"
    path: str = Field(description="Path to the DuckDB database file.")

    def get_ducklake_sql_parts(self, alias: str) -> Tuple[str, str]:
        """For file-based backends, no credential secret is needed."""
        return "", f"METADATA_PATH '{self.path}'"


class S3Config(dg.Config):
    """Configuration for an S3-compatible storage backend."""

    type: Literal["s3"] = "s3"
    endpoint_url: str = Field(
        description="Endpoint URL for the S3-compatible object store."
    )
    bucket: str = Field(description="Name of the S3 bucket for data storage.")
    prefix: str | None = Field(
        default=None, description="Optional path prefix within the S3 bucket."
    )
    aws_access_key_id: str
    aws_secret_access_key: str
    region: str = Field(default="us-east-1")
    use_ssl: bool = Field(default=True)
    url_style: str = Field(
        default="path", description="URL style for S3 ('path' or 'virtual')."
    )

    @property
    def full_data_path(self) -> str:
        """
        Handles the prefix gracefully. This is cleaner and more robust than a complex one-liner.
        It ensures a single trailing slash is present if a prefix is used.
        """
        path = f"s3://{self.bucket}"
        if self.prefix:
            clean_prefix = self.prefix.strip("/")
            return f"{path}/{clean_prefix}/"
        return f"{path}/"

    def get_ducklake_sql_parts(self, alias: str) -> Tuple[str, str]:
        """Returns the SQL for the credential secret and the main DATA_PATH parameter."""
        secret_name = f"secret_storage_{alias}"
        secret_sql = f"""
            CREATE OR REPLACE SECRET {secret_name} (
                TYPE S3, KEY_ID '{self.aws_access_key_id}',
                SECRET '{self.aws_secret_access_key}',
                ENDPOINT '{self.endpoint_url}', URL_STYLE '{self.url_style}',
                REGION '{self.region}', USE_SSL {"true" if self.use_ssl else "false"},
                SCOPE 's3://{self.bucket}'
            );
        """
        data_path_sql = f"DATA_PATH '{self.full_data_path}'"
        return secret_sql, data_path_sql


class DuckLakeLocalDirectory(dg.Config):
    """Configuration for a local filesystem storage directory."""

    type: Literal["local"] = "local"
    path: str = Field(description="Path to the local storage directory.")

    def get_ducklake_sql_parts(self, alias: str) -> Tuple[str, str]:
        """For local storage, no credential secret is needed."""
        return "", f"DATA_PATH '{self.path}'"


class DuckLakeResource(DuckDBConnectionProvider):
    """A highly configurable Dagster resource for interacting with DuckLake.

    Supports multiple metadata and storage backends.
    """

    metadata_backend: Union[PostgresConfig, SqliteConfig, DuckDBConfig] = Field(
        discriminator="type",
        description="Configuration for the metadata catalog backend.",
    )
    storage_backend: Union[S3Config, DuckLakeLocalDirectory] = Field(
        discriminator="type", description="Configuration for the data storage backend."
    )
    alias: str = Field(
        default="ducklake", description="Alias for the attached DuckLake instance."
    )
    plugins: List[str] = Field(
        default=["ducklake"],
        description="List of DuckDB plugins to install and load.",
    )

    def get_engine(self) -> Engine:
        engine = create_engine("duckdb:///:memory:")
        event.listen(engine, "connect", self._setup_ducklake_connection)
        return engine

    @contextmanager
    def connect(self) -> Generator[Connection, None, None]:
        """Yields a SQLAlchemy connection for use in a `with` statement."""
        with self.get_engine().connect() as conn:
            yield conn

    @contextmanager
    def duckdb_connect(self) -> Generator[DuckDBPyConnection, None, None]:
        """Yields a pre-configured native DuckDB connection for safe use."""
        conn = self.get_duckdb_connection()
        try:
            yield conn
        finally:
            conn.close()

    def get_duckdb_connection(self) -> DuckDBPyConnection:
        """Returns a new, pre-configured native duckdb connection object."""
        conn = duckdb.connect(database=":memory:")
        self._setup_ducklake_connection(conn, connection_record=None)
        return conn

    def _setup_ducklake_connection(self, dbapi_connection, connection_record):
        """
        Internal method to configure a DuckDB connection to use DuckLake.
        This now orchestrates SQL generation by calling its backend configs.
        """
        cursor = dbapi_connection.cursor()

        for plugin in self.plugins:
            cursor.execute(f"INSTALL {plugin}; LOAD {plugin};")

        metadata_secret_sql, metadata_params_sql = (
            self.metadata_backend.get_ducklake_sql_parts(self.alias)
        )
        storage_secret_sql, storage_data_path_sql = (
            self.storage_backend.get_ducklake_sql_parts(self.alias)
        )

        if metadata_secret_sql:
            cursor.execute(metadata_secret_sql)
        if storage_secret_sql:
            cursor.execute(storage_secret_sql)

        ducklake_secret_name = f"secret_{self.alias}"
        cursor.execute(f"""
            CREATE OR REPLACE SECRET {ducklake_secret_name} (
                TYPE DUCKLAKE,
                {metadata_params_sql},
                {storage_data_path_sql}
            );
        """)

        cursor.execute(f"ATTACH 'ducklake:{ducklake_secret_name}' AS {self.alias};")
        cursor.execute(f"USE {self.alias};")
        cursor.close()



ccc_version = "1.0.0"
step1_cc_index_meta_nodes = "nodes-meta"
step1_cc_index_nodes = "nodes"

step2_cc_edges = "edges"

step3_cc_graph = "graph"

step4_cc_graph_aggregated = "graph-aggregated-registered-domain"

step5_cc_tokenization = "tokenization"
step5_cc_frequency = "token-frequency"
step5_cc_models = "models"
step6_cc_cluster = "cluster-prediction"



from typing import Any, Dict

import dagster as dg
from dagster_dbt import DbtCliResource  # pyrefly: ignore


class DbtCliResourceWithConfig(dg.ConfigurableResource):
    dbt_project: DbtCliResource
    dbt_vars: Dict[str, Any]



import logging
import re
from typing import Optional
from urllib.parse import urljoin, urlparse

import cchardet
import idna
import tldextract
from bs4 import BeautifulSoup
from docling.datamodel.base_models import (
    InputFormat,
)
from docling.document_converter import DocumentConverter
from docling_core.types.doc import ImageRefMode

from ascii_library.cleanup.commoncrawl.text_extraction import TextExtractor

logging.basicConfig(
    level=logging.WARN, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)

# from surt import surt

_HTML_MIME = {"text/html"}
_XML_MIME = {
    "text/xml",
    "application/xml",
    "application/rss+xml",
    "application/atom+xml",
    "application/xhtml+xml",
    "application/xml-external-parsed-entity",
}


# Tags and their respective attributes that might contain links
standard_tags = {
    "a": "href",
    "link": "href",
    "img": "src",
    "script": "src",
    "object": "data",
    "iframe": "src",
    "source": "src",
    "track": "src",
    "area": "href",
    "base": "href",
    "blockquote": "cite",
    "q": "cite",
}

meta_tags = {
    "og:url",
    "og:image",
    "og:image:secure_url",
    "og:video",
    "og:video:url",
    "og:video:secure_url",
    "twitter:url",
    "twitter:image:src",
    "twitter:image",
    "thumbnail",
    "application-url",
    "msapplication-starturl",
    "msapplication-TileImage",
    "vb_meta_bburl",
}

# url_parse_host_pattern = re.compile(
#    r"^https?://([a-z0-9_.-]{2,253})(?:[/?#]|\Z)", re.IGNORECASE | re.ASCII
# )
# match IP addresses
# - including IPs with leading `www.' (stripped)
ip_pattern = re.compile(r"^(?:www\.)?\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\Z")


# valid host names, relaxed allowing underscore, allowing also IDNAs
# https://en.wikipedia.org/wiki/Hostname#Restrictions_on_valid_hostnames
host_part_pattern = re.compile(
    r"^[a-z0-9]([a-z0-9_-]{0,61}[a-z0-9])?\Z", re.IGNORECASE | re.ASCII
)

# pattern to split a data URL (<scheme>://<netloc>/<path> or <scheme>:/<path>)
data_url_pattern = re.compile("^(s3|s3a|https?|file|hdfs):(?://([^/]*))?/(.*)")


def get_surt_host(url):  # noqa: C901
    extracted = tldextract.extract(url, include_psl_private_domains=True)
    registered_domain = extracted.top_domain_under_public_suffix

    if registered_domain == "":
        registered_domain = f"{extracted.subdomain}.{extracted.domain}"
        if registered_domain == "":
            try:
                # Fallback to urlparse if tldextract fails
                host = urlparse(url).hostname
            except Exception as e:
                logger.debug(f"Failed to parse URL {url}: {e}")
                return None
            if not host:
                return None
        else:
            host = registered_domain
    else:
        host = registered_domain

    host = host.strip().lower()
    if len(host) < 1 or len(host) > 253:
        return None
    if ip_pattern.match(host):
        return None
    parts = host.split(".")
    if parts[-1] == "":
        # trailing dot is allowed, strip it
        parts = parts[0:-1]
    if len(parts) <= 1:
        # do not accept single-word hosts, must be at least `domain.tld'
        return None
    if len(parts) > 2 and parts[0] == "www":
        # strip leading 'www' to reduce number of "duplicate" hosts,
        # but leave at least 2 trailing parts (www.com is a valid domain)
        parts = parts[1:]
    for i, part in enumerate(parts):
        if len(part) > 63:
            return None
        if not host_part_pattern.match(part):
            try:
                idn = idna.encode(part).decode("ascii")
            except (
                idna.IDNAError,
                idna.core.InvalidCodepoint,
                UnicodeError,
                IndexError,
                Exception,
            ):
                logger.debug("Invalid host name: {}".format(url))
                return None

            # TODO: idna verifies the resulting string for length restrictions or invalid chars,
            #       maybe no further verification is required:
            if host_part_pattern.match(idn):
                parts[i] = idn
            else:
                logger.debug("Invalid host name: {}".format(url))
                return None
    parts.reverse()
    return ".".join(parts)


# get_surt_host('https://google.com')


def get_surt_host_tld(url):
    # Use tldextract to get the registered domain
    extracted = tldextract.extract(url, include_psl_private_domains=True)
    registered_domain = extracted.top_domain_under_public_suffix

    # Form the src_url_surt_host using the registered domain
    surt_host = f"https://{registered_domain}"
    return surt_host


def compute_surt_host(src_url, url):
    u = urljoin(src_url, url)
    # res = get_surt_host(u)
    res = get_surt_host(
        f"https://{tldextract.extract(u, include_psl_private_domains=True).top_domain_under_public_suffix}"
    )
    return res


def _normalize_mime(ct: Optional[str]) -> Optional[str]:
    if not ct:
        return None
    return ct.split(";")[0].strip().lower()


def _sniff_mode_from_head(content) -> str:  # noqa: C901
    if isinstance(content, bytes):
        head = content[:2048]
        # schnelle, verlustarme Decodierung für Sniffing
        try:
            h = head.decode("utf-8", "ignore")
        except Exception:
            h = head.decode("latin-1", "ignore")
    else:
        h = str(content)[:2048]
    h = h.strip().lower()

    if h.startswith(("<?xml", "<!xml", "<urlset", "<sitemapindex", "<rss", "<feed")):
        return "xml"
    if h.startswith(("<html", "<!doctype")):
        return "html"
    if "xmlns=" in h:
        return "xml"
    return "html"


def determine_content_type(record, content, content_type):  # noqa: C901
    # Check HTTP headers for Content-Type
    if content_type:
        mime = _normalize_mime(content_type)
        if mime in _HTML_MIME:
            return "lxml"
        if mime in _XML_MIME:
            return "lxml-xml"

    url = None
    try:
        url = record.headers.get("WARC-Target-URI")
    except Exception:
        url = None

    if url:
        lu = url.lower()
        if lu.endswith(
            (
                ".xml",
                ".rss",
                ".atom",
                ".xhtml",
                ".xht",
                ".rdf",
                ".opml",
                ".wsdl",
                ".xsd",
                ".svg",
            )
        ):
            return "lxml-xml"
        if lu.endswith((".html", ".htm", ".shtml", ".php", ".asp", ".aspx", ".jsp")):
            return "lxml"

    mode = _sniff_mode_from_head(content)
    return "lxml-xml" if mode == "xml" else "lxml"


def extract_charset_from_headers(headers):
    """Extract charset from the Content-Type header."""
    content_type = headers.get("Content-Type")
    if content_type:
        match = re.search(r"charset=([\w-]+)", content_type)
        if match:
            return match.group(1)
    return None


def detect_encoding(content_bytes):
    """Detect the encoding of content using chardet."""
    return cchardet.detect(content_bytes)["encoding"]


def should_parse_content_type(content_type):
    # Split on ";" and take the first part to get the MIME type only
    mime_type = content_type.split(";")[0].strip()
    # Check if the MIME type is one of the specified types
    return mime_type in ["text/html", "application/xml", "text/xml"]


def seems_like_binary(content):
    # Ensure content is treated as bytes
    if isinstance(content, str):
        content_bytes = content.encode()  # Convert to bytes
    else:
        content_bytes = content  # Assume it's already bytes

    return b"\x00" in content_bytes[:1024]  # Check the first 1KB for null byte


def extract_links_and_text_from_html(  # noqa: C901
    record, html_content, src_url, converter: DocumentConverter, payload_bytes
):  # noqa: C901
    """Extracts all links and their attributes from an HTML content using BeautifulSoup."""
    http_headers = record.http_headers
    if http_headers:
        content_type = http_headers.get("Content-Type")
    else:
        content_type = None

    link_data = []

    if not seems_like_binary(html_content) and should_parse_content_type(content_type):
        ct = determine_content_type(record, html_content, content_type)
        soup = BeautifulSoup(html_content, ct)

        def is_hidden(element):
            style = element.attrs.get("style", "")
            if "display: none" in style or "visibility: hidden" in style:
                return True
            if "position: absolute" in style and any(
                coord in style
                for coord in ["left: -", "top: -", "right: -", "bottom: -"]
            ):
                return True
            if element.has_attr("hidden"):  # Check for the "hidden" attribute
                return True
            return False

        for tag, attr in standard_tags.items():
            for t in soup.find_all(tag, **{attr: True}):  # pyrefly: ignore
                if is_hidden(t):
                    continue  # Skip hidden elements
                link_url = t[attr]  # pyrefly: ignore
                link_attributes = {
                    "url": str(link_url),
                    "surt_host": str(compute_surt_host(src_url, link_url)),
                    "attributes": {
                        str(k): str(v)
                        for k, v in t.attrs.items()  # pyrefly: ignore
                        if k != attr  # pyrefly: ignore
                    },
                }
                link_data.append(link_attributes)

        # Handle the meta tags
        for meta_tag in meta_tags:
            for t in soup.find_all(
                "meta", attrs={"property": meta_tag, "content": True}
            ):
                link_data.append(
                    {
                        "url": str(t["content"]),  # pyrefly: ignore
                        "surt_host": str(
                            compute_surt_host(src_url, t["content"])  # pyrefly: ignore
                        ),  # pyrefly: ignore
                        "attributes": {"property": str(meta_tag)},
                    }
                )
            for t in soup.find_all("meta", attrs={"name": meta_tag, "content": True}):
                link_data.append(
                    {
                        "url": str(t["content"]),  # pyrefly: ignore
                        "surt_host": str(
                            compute_surt_host(src_url, t["content"])  # pyrefly: ignore
                        ),  # pyrefly: ignore
                        "attributes": {"name": str(meta_tag)},
                    }
                )

        # Now extract the text using the TextExtractor, using the same soup
        # TODO: do the exception handling already here - is this causing the many logs?
        # try:
        text_extractor = TextExtractor()
        text_extractor.handle_node(soup)
        text_extractor.get_extracted_text()
        extracted_text = text_extractor.extracted_text  # None if no text extracted
        # except Exception:
        #    extracted_text = ""

        # TODO: When docling is enabled - things seem to be quite slow 15 minutes for the first 36 tasks of the dummy data!!!
        document_uri = record.headers.get("WARC-Target-URI")
        try:
            # WARNING: Silently ignores exeptions - we do not want to fail the whole job because of docling issues
            # possibly cut off here
            # max_num_pages: int = sys.maxsize,
            # max_file_size: int = sys.maxsize,
            res = converter.convert_string(
                html_content, format=InputFormat.HTML, name="inline.html"
            )
            doc = res.document
            extracted_text_clean = doc.export_to_markdown(
                image_mode=ImageRefMode.PLACEHOLDER
            )
        except Exception:
            extracted_text_clean = ""
            logger.error(
                f"Docling failed to extract text from HTML content for: {document_uri}",
                exc_info=False,
            )

    return link_data, extracted_text, extracted_text_clean  # type: ignore


def extract_links_from_headers(headers, src_url):
    """Extracts links from HTTP headers and formats them to match the output schema of extract_links_from_html."""
    links_with_attributes = []

    # 1. Extract from Location header (commonly used for redirects)
    location = headers.get("Location")
    if location:
        links_with_attributes.append(
            {
                "url": location,
                "surt_host": compute_surt_host(src_url, location),
                "attributes": {"rel": "location"},
            }
        )

    # 2. Extract from Link header
    link_header = headers.get("Link")
    if link_header:
        # Link headers can contain multiple links separated by commas.
        # This regex captures URLs wrapped in < > and their associated attributes.
        link_pattern = re.compile(r'<([^>]+)>\s*;\s*rel="([^"]+)"')
        for match in link_pattern.findall(link_header):
            url, rel = match
            links_with_attributes.append(
                {
                    "url": url,
                    "surt_host": compute_surt_host(src_url, url),
                    "attributes": {"rel": rel},
                }
            )

    # 3. Content-Location header
    content_location = headers.get("Content-Location")
    if content_location:
        links_with_attributes.append(
            {
                "url": content_location,
                "surt_host": compute_surt_host(src_url, content_location),
                "attributes": {"rel": "content-location"},
            }
        )

    return links_with_attributes



import os

import dagster as dg


def get_env(deployment_key: str = "DAGSTER_DEPLOYMENT", default_value="dev"):
    if (
        (os.getenv("DAGSTER_CLOUD_IS_BRANCH_DEPLOYMENT", "") == "1")
        or (os.environ.get(deployment_key, default_value) == "dev")
        or (os.environ.get(deployment_key, default_value) == "BRANCH")
    ):
        return "BRANCH"
    if (
        (os.getenv("DAGSTER_CLOUD_DEPLOYMENT_NAME", "") == "prod")
        or (os.environ.get(deployment_key, default_value) == "prod")
        or (os.environ.get(deployment_key, default_value) == "PROD")
    ):
        return "PROD"
    elif os.environ.get(deployment_key, default_value) == "INTEGRATION_TEST":
        return "INTEGRATION_TEST"
    raise ValueError(
        f"Unknown environment: {os.environ.get(deployment_key, default_value)}"
    )


def get_dagster_deployment_environment(
    deployment_key: str = "DAGSTER_DEPLOYMENT", default_value="dev"
):
    deplyoment = get_env(deployment_key, default_value)
    dg.get_dagster_logger().debug("dagster deployment environment: %s", deplyoment)
    return deplyoment



import os
import shlex
import signal
from io import BytesIO, StringIO
from textwrap import dedent
from types import FrameType
from typing import Any, Dict, List, Optional, Tuple

from dagster import (
    OpExecutionContext,
    PipesContextInjector,
    PipesMessageReader,
    ResourceParam,
    get_dagster_logger,
    open_pipes_session,
)
from dagster._core.pipes.client import PipesClientCompletedInvocation  # type: ignore
from dagster_aws.pipes import PipesS3ContextInjector, PipesS3MessageReader
from dagster_pipes import PipesExtras

from ascii_library.orchestration.pipes import LibraryConfig, LibraryKind
from ascii_library.orchestration.pipes.cloud_client import _PipesBaseCloudClient
from ascii_library.orchestration.pipes.exceptions import CustomPipesException
from ascii_library.orchestration.pipes.instance_config import CloudInstanceConfig
from ascii_library.orchestration.pipes.utils import (
    library_from_dbfs_paths,
    library_to_cloud_paths,
)
from ascii_library.orchestration.resources.constants import aws_region, rackspace_user
from ascii_library.orchestration.resources.emr_constants import pipeline_bucket
from ascii_library.utils.determine_env import (
    get_dagster_deployment_environment,
)

TARGET_DIR = "/mnt/pydeps/ascii_extras"
LOG_FILE = "/var/log/emr-bootstrap.log"


class _PipesEmrClient(_PipesBaseCloudClient):
    """Pipes client for EMR.

    Args:
        emr_job_runner (EmrJobRunner): An instance of EmrJobRunner.
        env (Optional[Mapping[str, str]]): An optional dict of environment variables to pass to the EMR job.
        context_injector (Optional[PipesContextInjector]): A context injector to use to inject context into the EMR process.
        message_reader (Optional[PipesMessageReader]): A message reader to use to read messages from the EMR job.
        poll_interval_seconds (float): How long to sleep between checking the status of the job run.
    """

    def __init__(
        self,
        emr_client,
        s3_client,
        price_client,
        bucket: str,
        context_injector: Optional[PipesContextInjector] = None,
        message_reader: Optional[PipesMessageReader] = None,
    ):
        super().__init__(
            main_client=emr_client,
            context_injector=context_injector,
            message_reader=message_reader,
            s3_client=s3_client,
        )
        self._price_client = price_client
        self._emr_client = emr_client
        self._s3_client = s3_client

        self._context_injector = context_injector or PipesS3ContextInjector(
            bucket=bucket, client=s3_client
        )
        self._message_reader = message_reader or PipesS3MessageReader(
            bucket=bucket, client=s3_client
        )

    def write_block(self, content, s: str):
        content.write(dedent(s).lstrip())

    def create_bootstrap_script(  # noqa: C901
        self,
        output_file: str = "bootstrap.sh",
        bucket: str = pipeline_bucket,
        libraries: Optional[List[LibraryConfig]] = None,
    ):
        dagster_deployment = get_dagster_deployment_environment()
        content = StringIO()
        self.write_block(
            content,
            """
                #!/bin/bash
                set -euo pipefail
                """,
        )
        # TODO: Fix eventually and move to pixi controlled venv; this is a bit hacky
        if libraries is not None:
            self.write_block(
                content,
                f"""
                LOG="{LOG_FILE}"
                sudo mkdir -p "$(dirname "$LOG")"
                # log the whole script as root → no permission errors
                exec > >(sudo tee -a "$LOG") 2>&1
                trap 'echo "== $(date) bootstrap FAILED (rc=$?) ==" >&2' ERR
                echo "== $(date) bootstrap start =="

                # make pip quiet and suppress 'running as root' warning
                export PIP_DISABLE_PIP_VERSION_CHECK=1
                export PIP_ROOT_USER_ACTION=ignore

                TARGET="{TARGET_DIR}"
                sudo mkdir -p "$TARGET"
                sudo chmod -R a+rX "$TARGET"

                # base Python
                # sudo yum -y update
                # sudo yum -y install python3 python3-pip
                command -v /usr/bin/python3 >/dev/null 2>&1 || sudo yum -y install python3
                /usr/bin/python3 -m pip --version >/dev/null 2>&1 || sudo yum -y install python3-pip
                /usr/bin/python3 --version
                /usr/bin/python3 -m pip --version

                # make target visible to future shells (write LITERAL path)
                echo 'export PYTHONPATH="{TARGET_DIR}${{PYTHONPATH:+:$PYTHONPATH}}"' | sudo tee /etc/profile.d/ascii_extras_path.sh >/dev/null
                sudo chmod +x /etc/profile.d/ascii_extras_path.sh
                # also for this shell:
                export PYTHONPATH="{TARGET_DIR}${{PYTHONPATH:+:$PYTHONPATH}}"
            """,
            )
            # load pypi first so we can reference in wheel
            for kind in (LibraryKind.Pypi, LibraryKind.Wheel):
                for lib in libraries:
                    if lib.kind == kind:
                        if kind == LibraryKind.Pypi:
                            self.handle_pypi(content, lib)
                        else:
                            self.handle_wheel(bucket, content, lib)
        self.write_block(
            content,
            r"""
            SITEPKG="$(
            /usr/bin/python3 -c 'import site,sys;p=[x for x in site.getsitepackages() if x.endswith("site-packages")];sys.stdout.write(p[0] if p else "")' \
            2>/dev/null || echo ""
            )"
            if [ -n "$SITEPKG" ] && [ -d "$SITEPKG" ]; then
            if ! echo "{TARGET_DIR}" | sudo tee "$SITEPKG/zzz_ascii_extras.pth" >/dev/null; then
                echo "[WARN] Could not write zzz_ascii_extras.pth to $SITEPKG; relying on PYTHONPATH"
            fi
            fi
        """.replace("{TARGET_DIR}", TARGET_DIR),
        )

        destination = f"external_pipes/{dagster_deployment}/{output_file}"
        content.seek(0)
        get_dagster_logger().debug(f"Bootstrap file content: \n\n{content.getvalue()}")
        self._s3_client.upload_fileobj(
            BytesIO(content.read().encode()), bucket, destination
        )
        return f"s3://{bucket}/{destination}"

    def handle_pypi(self, content, lib):
        pkg = f"{lib.name_id}{lib.version or ''}"
        flags = (lib.extra_flags or "").strip()
        get_dagster_logger().debug(f"Installing library: {pkg}")

        # Build the pip command once (into TARGET, ignore RPM packages)
        base = (
            "sudo /usr/bin/python3 -m pip install "
            "--upgrade --ignore-installed --no-cache-dir "
            '--target "$TARGET"'
        )
        if flags:
            base += f" {flags}"

        cmd = f"{base} {shlex.quote(pkg)}"
        self.write_block(
            content,
            f"""
            # Install {pkg} into $TARGET (isolated; no RPM conflicts)
            {cmd}
        """,
        )

    def handle_wheel(self, bucket, content, lib):
        name_id = library_from_dbfs_paths(lib.name_id)
        path = library_to_cloud_paths(lib_name=name_id, filesystem="s3")
        local = f"/tmp/{name_id}-0.0.0-py3-none-any.whl"

        self.write_block(
            content,
            f"""
            # Wheel: s3://{bucket}/{path}
            aws s3 cp s3://{bucket}/{path} {shlex.quote(local)}
            sudo /usr/bin/python3 -m pip install --upgrade --ignore-installed --no-cache-dir --target "$TARGET" {shlex.quote(local)}
        """,
        )
        get_dagster_logger().debug(f"Installing library: {name_id}")

    def modify_env_var(self, cluster_config: dict, key: str, value: str):
        configs = cluster_config.get("Configurations", [])
        i = 0
        for config in configs:
            if config.get("Classification") == "spark-defaults":
                props = config.get("Properties")
                # props = config.get("Configurations")[0].get("Properties")
                props[f"spark.yarn.appMasterEnv.{key}"] = value
                props[f"spark.executorEnv.{key}"] = value
                cluster_config["Configurations"][i]["Properties"] = props
            i += 1
        return cluster_config

    def extract_filename_without_extension(self, path: str):
        # Extract the base name of the file
        base_name = os.path.basename(path)
        # Remove the extension
        name_without_extension = os.path.splitext(base_name)[0]
        return name_without_extension

    def prepare_emr_job(
        self,
        local_file_path: str,
        bucket: str,
        s3_path: str,
        emr_job_config: Dict[str, Any],
        step_config,
        libraries_to_build_and_upload: Optional[List[str]] = None,
        libraries: Optional[List[LibraryConfig]] = None,
        extras: Optional[PipesExtras] = None,
    ) -> Tuple[(Optional[PipesExtras], Dict[str, Any])]:
        self._upload_file_to_cloud(
            local_file_path=local_file_path, bucket=bucket, cloud_path=s3_path
        )
        if libraries_to_build_and_upload is not None:
            self._ensure_library_on_cloud(
                libraries_to_build_and_upload=libraries_to_build_and_upload
            )
            output_file_name = f"{self.extract_filename_without_extension(local_file_path)}_bootstrap.sh"
            destination = self.create_bootstrap_script(
                output_file=output_file_name, libraries=libraries
            )
            emr_job_config = dict(emr_job_config)
            emr_job_config["BootstrapActions"] = [
                {
                    "Name": "Install custom packages",
                    "ScriptBootstrapAction": {"Path": destination},
                }
            ]
        if extras:
            # Create a mutable copy of extras if it exists
            extras = dict(extras) if extras else {}
            # TODO: do we really have to cast? extras = dict(extras)
            extras["emr_job_config"] = emr_job_config
            extras["step_config"] = step_config
        return extras, emr_job_config

    def adjust_emr_job_config(
        self,
        emr_job_config: dict,
        fleet_config: Optional[CloudInstanceConfig],
    ) -> dict:
        if (
            emr_job_config["Instances"].get("InstanceGroups") is None
            and emr_job_config["Instances"].get("InstanceFleets") is None
        ):
            if fleet_config is not None:
                emr_job_config["Instances"]["InstanceFleets"] = (
                    fleet_config.get_fleet_programatically(
                        emrClient=self._emr_client, priceClient=self._price_client
                    )
                )
                emr_job_config["ManagedScalingPolicy"]["ComputeLimits"]["UnitType"] = (
                    "InstanceFleetUnits"
                )
                emr_job_config["Instances"]["Ec2SubnetId"] = ""
            else:
                raise ValueError(
                    "No instance groups or fleets defined, and fleet_config is None."
                )
        elif emr_job_config["Instances"].get("InstanceGroups") is not None:
            emr_job_config["Instances"]["Ec2SubnetIds"] = []
        return emr_job_config

    def submit_emr_job(
        self,
        bootstrap_env,
        emr_job_config: dict,
        step_config,
        extras: PipesExtras,
    ) -> str:
        # get_dagster_logger().debug(
        #     f"DAGSTER_PIPES_CONTEXT: {bootstrap_env['DAGSTER_PIPES_CONTEXT']}"
        # )
        # get_dagster_logger().debug(
        #     f"DAGSTER_PIPES_MESSAGES: {bootstrap_env['DAGSTER_PIPES_MESSAGES']}"
        # )
        emr_job_config = self.modify_env_var(
            cluster_config=emr_job_config,
            key="DAGSTER_PIPES_CONTEXT",
            value=bootstrap_env["DAGSTER_PIPES_CONTEXT"],
        )
        emr_job_config = self.modify_env_var(
            cluster_config=emr_job_config,
            key="DAGSTER_PIPES_MESSAGES",
            value=bootstrap_env["DAGSTER_PIPES_MESSAGES"],
        )
        ascii_wandb_value = os.environ.get("ASCII_WANDB", "")
        if not ascii_wandb_value:
            get_dagster_logger().warning(
                "Environment variable 'ASCII_WANDB' is not set; defaulting to empty value."
            )
        emr_job_config = self.modify_env_var(
            cluster_config=emr_job_config,
            key="ASCII_WANDB",
            value=ascii_wandb_value,
        )
        emr_job_config = self.modify_env_var(
            cluster_config=emr_job_config,
            key="PYTHONPATH",
            value=f"{TARGET_DIR}:${{PYTHONPATH:+:$PYTHONPATH}}",
        )

        job_flow = self._emr_client.run_job_flow(**emr_job_config)
        get_dagster_logger().debug(f"EMR configuration: {job_flow}")
        self._emr_client.add_tags(
            ResourceId=job_flow["JobFlowId"],
            Tags=[
                {"Key": "jobId", "Value": job_flow["JobFlowId"]},
                {"Key": "executionMode", "Value": extras["execution_mode"]},
                {"Key": "engine", "Value": extras["engine"]},
            ],
        )
        self._emr_client.add_job_flow_steps(
            JobFlowId=job_flow["JobFlowId"],
            Steps=[step_config],
        )
        get_dagster_logger().info(
            f"If not signed in on Rackspace, please do it now: https://manage.rackspace.com/aws/account/{rackspace_user}/consoleSignin"
        )
        get_dagster_logger().info(
            f"EMR URL: https://{aws_region}.console.aws.amazon.com/emr/home?region={aws_region}#/clusterDetails/{job_flow['JobFlowId']}"
        )
        return job_flow["JobFlowId"]

    def run(  # noqa: C901 # type: ignore
        self,
        context: OpExecutionContext,
        emr_job_config: dict,
        step_config,  # Change from 'dict' to 'StepConfigTypeDef'
        local_file_path: str,
        bucket: str,
        s3_path: str,
        libraries_to_build_and_upload: Optional[List[str]] = None,
        libraries: Optional[List[LibraryConfig]] = None,
        extras: Optional[PipesExtras] = None,
        fleet_config: Optional[CloudInstanceConfig] = None,
    ) -> PipesClientCompletedInvocation:
        """Synchronously execute an EMR job with the pipes protocol."""
        emr_job_config = self.adjust_emr_job_config(emr_job_config, fleet_config)
        extras, emr_job_config = self.prepare_emr_job(
            local_file_path=local_file_path,
            bucket=bucket,
            s3_path=s3_path,
            emr_job_config=emr_job_config,
            step_config=step_config,
            libraries_to_build_and_upload=libraries_to_build_and_upload,
            libraries=libraries,
            extras=extras,
        )

        if extras is None:
            raise ValueError("Extras cannot be None.")

        cluster_id: Optional[str] = None

        def _terminate_cluster_safely():
            """Best-effort termination that works in STARTING/BOOTSTRAPPING/RUNNING/WAITING."""
            if not cluster_id:
                return
            try:
                # Optional: tag the cluster to mark the source of termination
                try:
                    self._emr_client.add_tags(
                        ResourceId=cluster_id,
                        Tags=[{"Key": "canceled-by", "Value": "dagster"}],
                    )
                except Exception:
                    pass
                context.log.info(f"[pipes] terminating EMR cluster {cluster_id}")
                self._emr_client.terminate_job_flows(JobFlowIds=[cluster_id])
            except Exception as te:  # swallow; termination is best-effort
                context.log.warning(
                    f"[pipes] could not terminate EMR {cluster_id}: {te!r}"
                )

        old_sigterm = signal.getsignal(signal.SIGTERM)

        def _on_sigterm(signum: int, frame: Optional[FrameType]):
            _terminate_cluster_safely()
            signal.signal(signal.SIGTERM, old_sigterm)
            os.kill(os.getpid(), signal.SIGTERM)

        signal.signal(signal.SIGTERM, _on_sigterm)

        try:
            with open_pipes_session(
                context=context,
                message_reader=self._message_reader,
                context_injector=self._context_injector,
                extras=extras,
            ) as session:
                bootstrap_env = session.get_bootstrap_env_vars()
                emr_job_config = extras.get("emr_job_config")  # type: ignore
                cluster_id = self.submit_emr_job(
                    bootstrap_env=bootstrap_env,
                    emr_job_config=emr_job_config,
                    step_config=step_config,
                    extras=extras,
                )
                self._poll_till_success(cluster_id=cluster_id)
        except CustomPipesException:
            # EMR or steps failed → terminate cluster to avoid lingering capacity
            context.log.info("[pipes] EMR/step failure detected; terminating cluster.")
            _terminate_cluster_safely()
            raise
        except BaseException:
            # Dagster canceled / SIGTERM / KeyboardInterrupt / any runtime error
            context.log.info(
                "[pipes] Dagster side interrupted; terminating EMR cluster."
            )
            _terminate_cluster_safely()
            raise
        finally:
            try:
                signal.signal(signal.SIGTERM, old_sigterm)
            except Exception:
                pass
            get_dagster_logger().debug("finished")
        return PipesClientCompletedInvocation(session)


PipesEmrEnhancedClient = ResourceParam[_PipesEmrClient]



import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd
from dagster import (
    ConfigurableIOManager,
    InputContext,
    OutputContext,
    ResourceDependency,
)
from dagster_pyspark.resources import LazyPySparkResource
from dagster_shared.check.functions import failed
from pyspark.sql import DataFrame as PySparkDataFrame


class PartitionedParquetIOManager(ConfigurableIOManager):  # type: ignore
    """This IOManager will take in a pandas or pyspark dataframe and store it in parquet at the
    specified path.

    It stores outputs for different partitions in different filepaths.

    Downstream ops can either load this dataframe into a spark session or simply retrieve a path
    to where the data is stored.
    """

    # Based on: https://github.com/dagster-io/dagster/blob/master/examples/project_fully_featured/project_fully_featured/resources/parquet_io_manager.py#L17

    _base_path: Optional[str]
    pyspark: ResourceDependency[LazyPySparkResource]
    _storage_options: Dict[str, Any]

    @property
    def storage_options(self):
        raise NotImplementedError()

    def handle_output(
        self, context: OutputContext, obj: Union[pd.DataFrame, PySparkDataFrame]
    ):
        path = self._get_path(context)
        context.log.debug(
            f"partitions: {context.has_asset_partitions}, type: {context.dagster_type.typing_type} ***"
        )
        if self._base_path is not None and "://" not in self._base_path:
            os.makedirs(os.path.dirname(path), exist_ok=True)

        if isinstance(obj, pd.DataFrame):
            row_count = len(obj)
            context.log.info(f"Row count: {row_count}")
            obj.to_parquet(
                path=path,
                index=False,
                compression="gzip",
                storage_options=self._storage_options,
            )
        elif isinstance(obj, PySparkDataFrame):
            obj.write.parquet(path=path, mode="overwrite", compression="gzip")
            row_count = self.pyspark.spark_session.read.parquet(path).count()
        else:
            raise Exception(f"Outputs of type {type(obj)} not supported.")

        context.add_output_metadata({"row_count": row_count, "path": path})

    def load_input(self, context) -> Union[PySparkDataFrame, pd.DataFrame, str]:
        path = self._get_path(context)
        context.log.debug(
            f"Trying to load input from {path} for type: {context.dagster_type.typing_type}"
        )
        if context.dagster_type.typing_type == PySparkDataFrame:
            return self.pyspark.spark_session.read.parquet(path)
        if context.dagster_type.typing_type == pd.DataFrame:
            return pd.read_parquet(path, storage_options=self._storage_options)
        if context.dagster_type.typing_type is str:
            return path
        return failed(
            f"Inputs of type {context.dagster_type} not supported. Please specify a valid type "
            "for this input either on the argument of the @asset-decorated function."
        )

    def _get_path(self, context: Union[InputContext, OutputContext]):
        key = context.asset_key.path[-1]

        if context.has_asset_partitions:
            start, end = context.asset_partitions_time_window
            dt_format = "%Y%m%d%H%M%S"
            # TODO potentially use hive-style partitioning for better spark support (if needed)
            partition_str = start.strftime(dt_format) + "_" + end.strftime(dt_format)
            if self._base_path is not None:
                return os.path.join(self._base_path, key, f"{partition_str}.parquet")
            else:
                return os.path.join(key, f"{partition_str}.parquet")
        else:
            if self._base_path is not None:
                return os.path.join(self._base_path, f"{key}.parquet")
            else:
                return f"{key}.parquet"


class LocalPartitionedParquetIOManager(PartitionedParquetIOManager):  # type: ignore
    base_key: Optional[str]

    @property  # type: ignore
    def _base_path(self):  # pyrefly: ignore
        out_path = Path("z_state") / "object_warehouse"
        out_path.mkdir(parents=True, exist_ok=True)

        if self.base_key:
            return os.path.join(str(out_path), self.base_key)
        else:
            return str(out_path)

    @property  # type: ignore
    def _storage_options(self):  # pyrefly: ignore
        return {}


class S3PartitionedParquetIOManager(PartitionedParquetIOManager):  # type: ignore
    s3_bucket: str
    access_key_id: str
    access_key_secret: str
    endpoint_url: Optional[str]

    @property  # type: ignore
    def _base_path(self):  # pyrefly: ignore
        return "s3a://" + self.s3_bucket

    @property  # type: ignore
    def _storage_options(self):  # pyrefly: ignore
        # Prep branch deployments:
        # listen to environment variables like:
        # DAGSTER_CLOUD_DEPLOYMENT_NAME for branch name and another one for boolean
        # provide configuration to overwrite - manually set branch (for testing)
        # cleanup: use S3 with short retention period
        # figure out how to generalize. Review also: https://docs.lakefs.io/
        # they will need to provide a cleanup hook
        so: Dict[str, Any] = {
            "key": self.access_key_id,
            "secret": self.access_key_secret,
        }
        if self.endpoint_url:
            so["client_kwargs"] = {}
            so["client_kwargs"]["endpoint_url"] = self.endpoint_url
        return so



import logging
import os
from io import BytesIO
from tempfile import SpooledTemporaryFile, TemporaryFile

import botocore
import requests
from docling.document_converter import DocumentConverter
from fastwarc.stream_io import FastWARCError
from fastwarc.warc import ArchiveIterator as FastWarcArchiveIterator
from fastwarc.warc import WarcRecord, WarcRecordType

from ascii_library.cleanup.commoncrawl.urls import data_url_pattern

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


def get_payload_stream(record: WarcRecord):
    return record.reader


def get_warc_header(
    record: WarcRecord,
    header: str,
    default: str = None,  # pyrefly: ignore
):  # pyrefly: ignore
    return record.headers.get(header, default)


def get_http_headers(record: WarcRecord):
    return record.http_headers.astuples()  # type: ignore


def is_response_record(record: WarcRecord):
    """Return true if WARC record is a WARC response record"""
    return record.record_type == WarcRecordType.response


def is_wet_text_record(record: WarcRecord):
    """Return true if WARC record is a WET text/plain record"""
    return (
        record.record_type == WarcRecordType.conversion
        and record.headers.get("Content-Type") == "text/plain"
    )


def is_wat_json_record(record: WarcRecord):
    """Return true if WARC record is a WAT record"""
    return (
        record.record_type == WarcRecordType.metadata
        and record.headers.get("Content-Type") == "application/json"
    )


def is_html(record: WarcRecord):  # noqa: C901
    """Return true if (detected) MIME type of a record is HTML"""
    html_types = ["text/html", "application/xhtml+xml"]
    if ("WARC-Identified-Payload-Type" in record.headers) and (
        record.headers["WARC-Identified-Payload-Type"] in html_types
    ):
        return True
    for name, value in record.http_headers.astuples():  # type: ignore
        if name.lower() == "content-type":
            content_type = value.lower()
            for html_type in html_types:
                if html_type in content_type:
                    return True
    return False


def fetch_warc(  # noqa: C901
    uri, warc_input_failed, base_uri=None, offset=-1, length=-1, s3client=None
):
    """Fetch WARC/WAT/WET files (or a record if offset and length are given)"""
    # s3client = boto3.client('s3', use_ssl=False)

    (scheme, netloc, path) = (None, None, None)
    uri_match = data_url_pattern.match(uri)
    if not uri_match and base_uri:
        # relative input URI (path) and base URI defined
        uri = base_uri + uri
        uri_match = data_url_pattern.match(uri)

    if uri_match:
        (scheme, netloc, path) = uri_match.groups()
    else:
        # keep local file paths as is
        path = uri

    stream = None

    if scheme in ["s3", "s3a"]:
        bucketname = netloc
        if not bucketname:
            logger.error("Invalid S3 URI: " + uri)
            return
        if not path:
            logger.error("Empty S3 path: " + uri)
            return
        elif path[0] == "/":
            # must strip leading / in S3 path
            path = path[1:]
        if offset > -1 and length > 0:
            rangereq = "bytes={}-{}".format(offset, (offset + length - 1))
            # Note: avoid logging too many small fetches
            logger.debug("Fetching {} ({})".format(uri, rangereq))
            try:
                response = s3client.get_object(  # pyrefly: ignore
                    Bucket=bucketname, Key=path, Range=rangereq
                )
                stream = BytesIO(response["Body"].read())
            except botocore.client.ClientError as e:  # type: ignore
                logger.error(
                    "Failed to download: s3://{}/{} (offset: {}, length: {}) - {}".format(
                        bucketname, path, offset, length, e
                    )
                )
                warc_input_failed.add(1)
                return
        else:
            logger.debug("Reading from S3 {}".format(uri))
            # download entire file using a temporary file for buffering
            warctemp = TemporaryFile(mode="w+b")  # , dir=args.local_temp_dir)
            # print('***************')
            # print(warctemp.name)
            try:
                s3client.download_fileobj(bucketname, path, warctemp)  # pyrefly: ignore
                warctemp.seek(0)
                stream = warctemp
                # stream = input_data
                # stream = FastWarcArchiveIterator(open(input_data, 'rb'))
            except botocore.client.ClientError as e:  # type: ignore
                logger.error("Failed to download {}: {}".format(uri, e))
                warc_input_failed.add(1)
                warctemp.close()

    elif scheme == "http" or scheme == "https":
        headers = None
        if offset > -1 and length > 0:
            headers = {"Range": "bytes={}-{}".format(offset, (offset + length - 1))}
            # Note: avoid logging many small fetches
            logger.debug("Fetching {} ({})".format(uri, headers))
        else:
            logger.debug("Fetching {}".format(uri))
            pass
        response = requests.get(uri, headers=headers)

        if response.ok:
            # includes "HTTP 206 Partial Content" for range requests
            warctemp = SpooledTemporaryFile(
                max_size=2097152,
                mode="w+b",
                # dir=args.local_temp_dir
            )
            warctemp.write(response.content)
            warctemp.seek(0)
            stream = warctemp
        else:
            logger.error("Failed to download {}: {}".format(uri, response.status_code))
            pass

    elif scheme == "hdfs":
        try:
            import pydoop.hdfs as hdfs  # type: ignore

            logger.error("Reading from HDFS {}".format(uri))
            stream = hdfs.open(uri)
        except RuntimeError as e:
            logger.error("Failed to open {}: {}".format(uri, e))
            warc_input_failed.add(1)

    else:
        logger.debug("Reading local file {}".format(uri))
        if scheme == "file":
            # must be an absolute path
            uri = os.path.join("/", path)
        else:
            base_dir = os.path.abspath(os.path.dirname(__file__))
            uri = os.path.join(base_dir, uri)
        try:
            stream = open(uri, "rb")
        except IOError as e:
            logger.error("Failed to open {}: {}".format(uri, e))
            warc_input_failed.add(1)

    return stream


def iterate_records(
    warc_uri,
    archive_iterator,
    process_record,
    records_processed,
    converter: DocumentConverter,
):
    """Iterate over all WARC records and process them"""
    for record in archive_iterator:
        for res in process_record(record, converter):
            yield res
        records_processed.add(1)


def process_warc(
    uri,
    stream,
    warc_input_failed,
    process_record,
    records_processed,
    converter: DocumentConverter,
):
    """Parse a WARC (or WAT/WET file) via FastWARC"""
    # process only WARC response and metadata (including WAT) records
    # WarcRecordType.metadata |
    fastwarc_record_filter = WarcRecordType.response

    warc_parse_http_header = True
    try:
        rec_iter = FastWarcArchiveIterator(
            stream,
            record_types=fastwarc_record_filter,
            parse_http=warc_parse_http_header,
        )
        for res in iterate_records(
            uri,
            rec_iter,
            process_record=process_record,
            records_processed=records_processed,
            converter=converter,
        ):
            yield res
    except FastWARCError as e:
        warc_input_failed.add(1)
        logger.error("Invalid WARC: {} - {}".format(uri, e))



import polars as pl
import tldextract

from ascii_library.cleanup.commoncrawl.urls import get_surt_host


def perform_extract(x):
    if x is not None:
        extracted = tldextract.extract(x, include_psl_private_domains=True)
        registered_domain = extracted.top_domain_under_public_suffix
        if registered_domain == "":
            registered_domain = f"{extracted.subdomain}.{extracted.domain}"
        return registered_domain
    else:
        return None


def clean_urls(combined_seeds: pl.DataFrame):
    """Input: Ploars dataframe with column: website_address"""
    return (
        combined_seeds.with_columns(
            pl.col("website_address")
            .str.to_lowercase()
            .str.replace("www..", "", literal=True)
            .str.replace("http:", "", literal=True)
            .str.replace("https:", "", literal=True)
            .str.replace("www.", "", literal=True)
            .str.replace("..", "", literal=True)
            .str.strip_chars("/")
            .str.split("/")
            .list.get(0)
            .str.replace("szmaxun.con", "szmaxun.com", literal=True)
            .alias("seed_node_url")
        )
        .with_columns(
            pl.col("seed_node_url")
            .map_elements(
                lambda x: (
                    tldextract.extract(
                        x, include_psl_private_domains=True
                    ).top_domain_under_public_suffix
                    if x is not None
                    else None
                ),
                return_dtype=pl.String,
            )
            .alias("extracted")
        )
        .filter(
            (pl.col("seed_node_url").str.len_bytes() >= 4)
            & (pl.col("seed_node_url").is_not_null())
        )
        .with_columns(
            pl.col("extracted")
            .map_elements(
                lambda x: get_surt_host(perform_extract(f"https://{x}"))
                if x is not None
                else None,
                return_dtype=pl.String,
            )
            .alias("seed_node_url_surt")
        )
    )


def filter_sane_urls(combined_seeds: pl.DataFrame):
    """Input: Ploars dataframe of shape: ascii_id_company, website_address"""
    df = clean_urls(combined_seeds)

    # the following nodes would be deleted
    to_remove = df.filter(
        pl.col("extracted") != pl.col("seed_node_url"),
        ~pl.col("extracted").is_in(["1688.com"]),
        pl.col("extracted").str.len_bytes() < 4,
    ).to_pandas()
    if to_remove.shape[0] > 0:
        print(f"Removing the following weird URLs: {to_remove.shape}")
        print(to_remove)
        print(f"before removal: {df.shape}")

    df_seeds_duplicated = df.filter(
        ~pl.col("extracted").is_in(["1688.com"]),
        pl.col("extracted").str.len_bytes() >= 4,
        pl.col("extracted").is_not_null(),
    ).select(
        pl.col("ascii_id_company"),
        pl.col("extracted").alias("seed_node_url"),
        pl.col("seed_node_url_surt"),
    )
    print(f"df_seeds_duplicated: {df_seeds_duplicated.shape}")

    df_seeds_production = df_seeds_duplicated.unique(subset=["seed_node_url"])
    print(f"df_seeds_production: {df_seeds_production.shape}")
    return df_seeds_production, df_seeds_duplicated, to_remove


def upload_seeds_to_s3(
    fs,
    filename: str,
    df_seeds_production: pl.DataFrame,
    df_seeds_duplicated: pl.DataFrame,
    BUCKET_NAME: str = "ascii-supply-chain-research-input",
):
    with fs.open(
        f"{BUCKET_NAME}/ascii_seeds/seeds_deduplicated/seeds={filename}/seeds.parquet",
        mode="wb",
    ) as f:
        df_seeds_production.write_parquet(f, compression="gzip")
    with fs.open(
        f"{BUCKET_NAME}/ascii_seeds/seeds_not_deduplicated/seeds={filename}/seeds_duplicated.parquet",
        mode="wb",
    ) as f:
        df_seeds_duplicated.write_parquet(f, compression="gzip")



from typing import Dict, Mapping, Optional, Sequence

from ascii_library.orchestration.pipes.instance_config import CloudInstanceConfig
from ascii_library.orchestration.resources import constants, emr_constants

from .constants import (
    job_role,
    master_security_group,
    service_role,
    slave_security_group,
)


def add_spark_settings(
    spark_settings: Dict[str, dict], additional_spark_args: Optional[Dict[str, str]]
) -> Dict[str, dict]:
    """Helper function to add additional Spark settings."""
    if additional_spark_args:
        for key, value in additional_spark_args.items():
            spark_settings["Properties"][key] = value
    return spark_settings or {}


def get_basic_config(name, s3_log_uri, job_role, service_role):
    """Create the basic cluster configuration."""
    return {
        "Name": name,
        "LogUri": s3_log_uri,
        "ReleaseLabel": emr_constants.releaseLabel,
        "Applications": [{"Name": "Spark"}],
        "VisibleToAllUsers": True,
        "JobFlowRole": job_role,
        "ServiceRole": service_role,
        "Tags": [{"Key": "for-use-with-amazon-emr-managed-policies", "Value": "true"}],
        "ScaleDownBehavior": "TERMINATE_AT_TASK_COMPLETION",
        "AutoTerminationPolicy": {"IdleTimeout": 180},
    }


def get_emr_cluster_config(
    minimumCapacityUnits: int,
    maximumCapacityUnits: int,
    maximumOnDemandCapacityUnits: int,
    # we do not store anything to HDFS, want to limit core nodes to a minimum
    maximumCoreCapacityUnits: int = 1,
    s3_log_uri: Optional[str] = constants.s3_log_uri,
    subnet_id: Optional[str] = emr_constants.Subnets.usEast1b.value,
    master_security_group: str = master_security_group,
    slave_security_group: str = slave_security_group,
    service_role: str = service_role,
    job_role: str = job_role,
    additional_master_security_groups: Optional[list[str]] = None,
    additional_slave_security_groups: Optional[list[str]] = None,
    ssh: str = "emr_key_v2",
    name: Optional[str] = "unnamed",
    group: Optional[Sequence[CloudInstanceConfig]] = None,
    fleet: Optional[Sequence[CloudInstanceConfig]] = None,
    additional_spark_args: Optional[Mapping[str, str]] = None,
) -> Dict:
    if additional_slave_security_groups is None:
        additional_slave_security_groups = [
            master_security_group,
            slave_security_group,
        ]
    if additional_master_security_groups is None:
        additional_master_security_groups = [
            master_security_group,
            slave_security_group,
        ]
    spark_settings = {
        "Classification": "spark-defaults",
        "Properties": {
            "spark.sql.extensions": "io.delta.sql.DeltaSparkSessionExtension",
            "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            # "spark.databricks.delta.schema.autoMerge.enabled": "True",
            "spark.databricks.delta.schema.autoMerge.enabledOnWrite": "True",
            "spark.jars.repositories": "https://repo1.maven.org/maven2/",
            # "spark.jars.packages": "io.delta:delta-spark_2.12:3.3.2,org.apache.hadoop:hadoop-aws:3.3.4,org.apache.spark:spark-hadoop-cloud_2.12:3.5.0,com.johnsnowlabs.nlp:spark-nlp_2.12:5.3.3",
            # "spark.jars.packages": "io.delta:delta-spark_2.12:3.3.2,org.postgresql:postgresql:42.7.7,org.apache.hadoop:hadoop-aws:3.3.6,org.apache.spark:spark-hadoop-cloud_2.12:3.5.6,com.johnsnowlabs.nlp:spark-nlp_2.12:6.1.3",
            # AI Tells me that delta/storage jars should be there by default - lets try without them
            "spark.jars.packages": "org.postgresql:postgresql:42.7.7,com.johnsnowlabs.nlp:spark-nlp_2.12:6.1.3",
            "spark.hadoop.fs.s3a.s3guard.ddb.region": "us-east-1",
            "spark.submit.deployMode": "cluster",
            # not needed, these paramters are DIRECTLY managed by the compute provider by selecting an instance type
            # "spark.executor.memory": executorMemory,
            # "spark.driver.memory": driverMemory,
            "spark.driver.maxResultSize": "10G",
            # "spark.storage.memoryFraction": storageMemoryFraction,
            # "spark.shuffle.memoryFraction": "0.2",
            # hard coding here TODO pass in dynamically
            # "spark.driver.memory": "30G",
            # "spark.executor.memory": "8575m",
            # "spark.executor.memory": "33G",
            # "spark.executor.memory": "8G",
            # "spark.executor.memoryOverhead": "2G",
            "spark.sql.parquet.columnarReaderBatchSize": "2048",  # half of default
            # "spark.memory.offHeap.enabled": "true",
            # "spark.memory.offHeap.size": "24G",
            # "spark.memory.offHeap.size":36974886912,
            "spark.blacklist.decommissioning.timeout": "300s",
            "spark.yarn.appMasterEnv.SPARK_PIPES_ENGINE": "emr",
            "spark.yarn.maxAppAttempts": "1",
        },
    }
    yarn_settings = {
        "Classification": "yarn-site",
        "Properties": {
            "yarn.resourcemanager.nodemanager-graceful-decommission-timeout-secs": "300",
            "yarn.node-labels.enabled": "true",
            "yarn.node-labels.am.default-node-label-expression": "CORE",
        },
    }
    spark_settings = add_spark_settings(spark_settings, additional_spark_args)  # type: ignore
    basic = get_basic_config(name, s3_log_uri, job_role, service_role)

    basic["Configurations"] = [
        {
            "Classification": "delta-defaults",
            "Properties": {"delta.enabled": "true"},
        },
        spark_settings,
        {
            "Classification": "spark",
            "Properties": {"maximizeResourceAllocation": "true"},
        },
        yarn_settings,
    ]

    instances = {
        # "MasterInstanceType":
        # "SlaveInstanceType":
        "Ec2KeyName": ssh,
        #'Placement': {
        #    'AvailabilityZone': 'string',
        #    'AvailabilityZones': [
        #        'string',
        #    ]
        # },
        "KeepJobFlowAliveWhenNoSteps": False,
        "TerminationProtected": False,
        #'HadoopVersion': 'string',
        "Ec2SubnetId": subnet_id,
        "Ec2SubnetIds": [
            emr_constants.Subnets.usEast1a.value,
            emr_constants.Subnets.usEast1b.value,
            emr_constants.Subnets.usEast1c.value,
            emr_constants.Subnets.usEast1d.value,
            emr_constants.Subnets.usEast1e.value,
            emr_constants.Subnets.usEast1f.value,
        ],
        "EmrManagedMasterSecurityGroup": master_security_group,
        "EmrManagedSlaveSecurityGroup": slave_security_group,
        #'ServiceAccessSecurityGroup': 'string',
        "AdditionalMasterSecurityGroups": additional_master_security_groups,
        "AdditionalSlaveSecurityGroups": additional_slave_security_groups,
    }
    computeLimits = {
        "MinimumCapacityUnits": minimumCapacityUnits,
        "MaximumCapacityUnits": maximumCapacityUnits,
        "MaximumOnDemandCapacityUnits": maximumOnDemandCapacityUnits,
        "MaximumCoreCapacityUnits": maximumCoreCapacityUnits,
    }
    if group is not None:
        computeLimits["UnitType"] = "Instances"  # type: ignore
        instances["InstanceGroups"] = group  # pyrefly: ignore
        basic["AutoScalingRole"] = "EMR_AutoScaling_DefaultRole"
    elif fleet is not None:
        computeLimits["UnitType"] = "InstanceFleetUnits"  # type: ignore
        instances["InstanceFleets"] = fleet  # pyrefly: ignore
    basic["ManagedScalingPolicy"] = {"ComputeLimits": computeLimits}  # type: ignore
    basic["Instances"] = instances  # type: ignore
    return basic



import json
import re
from typing import Any, Dict, List, Optional, Sequence

from dagster import get_dagster_logger

from ascii_library.orchestration.resources import emr_constants


class CloudInstanceConfig:
    def __init__(self, **kwargs):
        self.instanceRole = kwargs.get("InstanceRole")
        self.market = kwargs.get("Market")
        self.name = kwargs.get("Name")
        self.instanceType = kwargs.get("InstanceType")
        self.workerInstanceType = kwargs.get(
            "WorkerInstanceType", kwargs.get("InstanceType")
        )

        self.bidPrice = kwargs.get("BidPrice")
        self.customAmiId = kwargs.get("CustomAmiId")
        self.sizeInGB = kwargs.get("SizeInGB", emr_constants.sizeInGB)
        self.volumeType = kwargs.get("VolumeType", emr_constants.volumeType)
        self.volumesPerInstance = kwargs.get(
            "VolumesPerInstance", emr_constants.volumesPerInstance
        )
        self.ebsOptimized = kwargs.get("EbsOptimized")
        self.instanceCount = kwargs.get("InstanceCount")
        self.masterInstanceCount = kwargs.get(
            "MasterInstanceCount", kwargs.get("InstanceCount", 1)
        )
        self.coreInstanceCount = kwargs.get(
            "CoreInstanceCount", kwargs.get("InstanceCount", 1)
        )
        self.taskInstanceCount = kwargs.get(
            "TaskInstanceCount", kwargs.get("InstanceCount", 1)
        )
        self.percentageOfOnDemandPrice = kwargs.get(
            "PercentageOfOnDemandPrice", emr_constants.percentageOfOnDemandPrice
        )
        self.reservationPreference = kwargs.get("ReservationPreference")
        self.allocationStrategy = kwargs.get("AllocationStrategy")
        self.timeoutAction = kwargs.get("TimeoutAction")
        self.timeoutDuration = kwargs.get(
            "TimeoutDuration", emr_constants.timeoutDuration
        )
        self.targetOnDemand = kwargs.get("TargetOnDemand")
        self.targetOnSpot = kwargs.get("TargetOnSpot")
        self.weightedCapacity = kwargs.get("WeightedCapacity")
        self.memoryGB = kwargs.get("MemoryGB")
        self.storageGB = kwargs.get("StorageGB")
        self.noEBS = kwargs.get("NoEBS", False)
        self.filters = kwargs.get("Filters")
        self.maximumCoreCapacityUnits = kwargs.get("MaximumCoreCapacityUnits", 10)
        self.maximumOnDemandCapacityUnits = kwargs.get(
            "MaximumOnDemandCapacityUnits", 10
        )
        self.maximumCapacityUnits = kwargs.get("MaximumCapacityUnits", 100)

    def __repr__(self):
        return (
            f"CloudInstanceConfig(instanceRole={self.instanceRole}, market={self.market}, name={self.name}, "
            f"instanceType={self.instanceType}, workerInstanceType={self.workerInstanceType}, bidPrice={self.bidPrice}, "
            f"customAmiId={self.customAmiId}, sizeInGB={self.sizeInGB}, volumeType={self.volumeType}, "
            f"volumesPerInstance={self.volumesPerInstance}, ebsOptimized={self.ebsOptimized}, instanceCount={self.instanceCount}, "
            f"masterInstanceCount={self.masterInstanceCount}, coreInstanceCount={self.coreInstanceCount}, "
            f"taskInstanceCount={self.taskInstanceCount}, reservationPreference={self.reservationPreference}, "
            f"allocationStrategy={self.allocationStrategy}, timeoutAction={self.timeoutAction}, timeoutDuration={self.timeoutDuration}, "
            f"targetOnDemand={self.targetOnDemand}, targetOnSpot={self.targetOnSpot}, weightedCapacity={self.weightedCapacity}, "
            f"memoryGB={self.memoryGB}, storageGB={self.storageGB}, noEBS={self.noEBS}, filters={self.filters})"
        )

    def __str__(self):
        return self.__repr__()

    def create_filters(self, instance, ops: str = "Linux", region: str = "us-east-1"):
        """
        create an array with the filters fo price
        :param insance: json that from of an available instances on EMR,
        :param ops: operative system, EMR only takes Linux
        :param region: the region we work, by default it's us-east-1
        :return:
        """
        return [
            {"Type": "TERM_MATCH", "Field": "instanceType", "Value": instance["Type"]},
            {"Type": "TERM_MATCH", "Field": "operatingSystem", "Value": ops},
            {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
            {
                "Type": "TERM_MATCH",
                "Field": "licenseModel",
                "Value": "No License required",
            },
            {"Type": "TERM_MATCH", "Field": "regionCode", "Value": region},
        ]

    def get_instance_price(self, prices):
        """
        parse json with prices for getting the price fo the right type
        :param prics: a list with the return of pricing client
        :return: float with the right price
        """
        for i in range(len(prices)):
            terms = json.loads(prices[i])["terms"]
            for sku, details in terms.items():
                if sku == "OnDemand":
                    priceDimension = details[list(details.keys())[0]]["priceDimensions"]
                    description = priceDimension[list(priceDimension.keys())[0]]
                    if (
                        description["unit"] == "Hrs"
                        and float(description["pricePerUnit"]["USD"]) > 0.0
                    ):
                        return float(description["pricePerUnit"]["USD"])

    def add_prefix_suffix_to_instances(self, instances):
        """
        Adds 'Prefix' and 'Suffix' to each instance dictionary based on the 'Type'.

        :param instances: List of instance dictionaries with a 'Type' key.
        :return: The original list of instances, modified to include 'Prefix' and 'Suffix'.
        """
        for instance in instances:
            type_match = re.split(r"\.\s*", instance["Type"])
            instance["Prefix"] = type_match[0]
            instance["Suffix"] = type_match[1]
        return instances

    def filter_by_numeric_value(self, instances, key, value):
        return [instance for instance in instances if instance.get(key, 0) >= value]

    def filter_by_exact_match(self, instances, key, value):
        return [instance for instance in instances if instance.get(key) == value]

    def filter_by_enum(
        self, instances, value, base=emr_constants.Suffix
    ):  # I think this method could be extended to other enums if needed
        target_index = base.index_of(value)
        if target_index != -1:  # Make sure the suffix provided is valid
            return [
                instance
                for instance in instances
                if base.index_of(instance.get("Suffix")) <= target_index
            ]

    def filter_instances(self, instances, **kwargs):
        """
        Filters instances based on given criteria. Numeric fields are compared for greater or equal values,
        while string fields must match exactly.

        :param instances: List of instance dictionaries.
        :param kwargs: Criteria for filtering (e.g., MemoryGB=16, VCPU=4, Prefix='m5').
        :return: List of filtered instances.
        """
        for key, value in kwargs.items():
            if isinstance(value, (int, float)) and key in [
                "MemoryGB",
                "VCPU",
                "StorageGB",
            ]:
                # Filter for numeric values (equal or greater)
                instances = self.filter_by_numeric_value(instances, key, value)
            elif isinstance(value, str) and key in ["InstanceFamilyId", "Prefix"]:
                # Filter for exact string match
                instances = self.filter_by_exact_match(instances, key, value)
            elif isinstance(value, str) and key == "Suffix":
                instances = self.filter_by_enum(instances=instances, value=value)
        return instances

    def get_instance_config(self):
        config = self._base_config()
        self._add_bid_price(config)
        self._add_custom_ami_id(config)
        self._add_ebs_configuration(config)
        return config

    def _base_config(self):
        return {
            "InstanceCount": self.instanceCount,
            "InstanceRole": self.instanceRole.value,  # type: ignore
            "InstanceType": self.instanceType,
            "Market": self.market.value,  # type: ignore
            "Name": self.name,
        }

    def _add_bid_price(self, config):
        if self.market == emr_constants.Market.spot and self.bidPrice is not None:
            config["BidPrice"] = self.bidPrice

    def _add_custom_ami_id(self, config, index: int = 0):
        if self.customAmiId[index] is not None:  # type: ignore
            config["CustomAmiId"] = self.customAmiId[index]  # type: ignore

    def _add_ebs_configuration(self, config):
        # if self._is_ebs_configuration_valid():
        ebsConfiguration = {"EbsOptimized": self.ebsOptimized}
        ebsBlockDeviceConfigs = self._build_ebs_block_device_config()
        ebsConfiguration["EbsBlockDeviceConfigs"] = ebsBlockDeviceConfigs
        config["EbsConfiguration"] = ebsConfiguration

    def _is_ebs_configuration_valid(self):
        return (
            len(self.sizeInGB) == len(self.volumeType) == len(self.volumesPerInstance)
        )

    def _build_ebs_block_device_config(self, index: int = 0):
        volume_size = (
            self.sizeInGB if isinstance(self.sizeInGB, int) else self.sizeInGB[index]
        )
        volume_type = (
            self.volumeType
            if isinstance(self.volumeType, emr_constants.VolumeType)
            else self.volumeType[index]
        )
        block = {
            "VolumeSpecification": {
                "SizeInGB": volume_size,
                "VolumeType": volume_type.value,
            }
        }
        block["VolumesPerInstance"] = self.volumesPerInstance[index]
        return [block]

    def gen_instance_type_config(self):
        instanceTypeConfigs = []
        ebsBlockDeviceConfigs = self._build_ebs_block_device_config()
        temp = {
            "EbsConfiguration": {
                "EbsBlockDeviceConfigs": ebsBlockDeviceConfigs[0],
                "EbsOptimized": self.ebsOptimized[0],  # type: ignore
            },
            "InstanceType": self.instanceType[0],  # type: ignore
            "WeightedCapacity": self.weightedCapacity[0],  # type: ignore
        }
        if self.percentageOfOnDemandPrice is not None:
            temp["BidPriceAsPercentageOfOnDemandPrice"] = self.percentageOfOnDemandPrice
        if self.bidPrice[0] is not None:  # type: ignore
            temp["BidPrice"] = self.bidPrice[0]  # type: ignore
        instanceTypeConfigs.append(temp)
        return instanceTypeConfigs

    def get_launch_spec(self):
        launchSpecifications = {}
        if self.reservationPreference is not None:
            launchSpecifications["OnDemandSpecification"] = {
                "AllocationStrategy": "lowest-price",  # it's literally the only valid value, that's why it's hardcoded
                "CapacityReservationOptions": {"CapacityReservationPreference": "open"},
            }

        launchSpecifications["SpotSpecification"] = {  # pyrefly: ignore
            "AllocationStrategy": "capacity-optimized",
            "TimeoutAction": "SWITCH_TO_ON_DEMAND",
            "TimeoutDurationMinutes": 5,
        }
        return launchSpecifications

    def get_fleet_config(self):
        instanceTypeConfigs = self.gen_instance_type_configs()
        fleet_config = {"InstanceFleetType": self.instanceRole, "Name": self.name}

        if self.targetOnDemand is not None:
            fleet_config["TargetOnDemandCapacity"] = self.targetOnDemand
        if self.targetOnSpot is not None:
            fleet_config["TargetSpotCapacity"] = self.targetOnSpot
        fleet_config["InstanceTypeConfigs"] = instanceTypeConfigs
        if self.timeoutDuration is not None:
            fleet_config["ResizeSpecifications"] = {
                "OnDemandResizeSpecification": {
                    "AllocationStrategy": "lowest-price",
                    "TimeoutDurationMinutes": self.timeoutDuration,
                },
                "SpotResizeSpecification": {
                    "TimeoutDurationMinutes": self.timeoutDuration
                },
            }
        fleet_config["LaunchSpecifications"] = self.get_launch_spec()
        return fleet_config

    def get_default_fleet(self, **kwargs):
        master_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.master.value,
            Market=emr_constants.Market.on_demand.value,
            Name="master",
            InstanceType=[self.instanceType],
            WeightedCapacity=[1],
            SizeInGB=[self.sizeInGB],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=[self.ebsOptimized],
            TargetOnDemand=1,
            ReservationPreference=self.reservationPreference,
            TimeoutDuration=self.timeoutDuration,
        )

        core_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.core.value,
            Market=emr_constants.Market.on_demand.value,
            Name="core",
            InstanceType=[kwargs.get("WorkerInstanceType", self.instanceType)],
            SizeInGB=[self.sizeInGB],
            WeightedCapacity=[self.weightedCapacity],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=[self.ebsOptimized],
            TargetOnDemand=self.targetOnDemand,
            ReservationPreference=self.reservationPreference,
            TimeoutDuration=self.timeoutDuration,
        )

        task_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.task.value,
            Market=emr_constants.Market.spot.value,
            Name="task",
            InstanceType=[kwargs.get("WorkerInstanceType", self.instanceType)],
            SizeInGB=[self.sizeInGB],
            WeightedCapacity=[self.weightedCapacity],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=[self.ebsOptimized],
            PercentageOfOnDemandPrice=self.percentageOfOnDemandPrice,
            TimeoutAction=self.timeoutAction,
            ReservationPreference=self.reservationPreference,
            TimeoutDuration=self.timeoutDuration,
            TargetOnDemand=0,
            TargetOnSpot=self.targetOnSpot,
        )

        return [
            master_config.get_fleet_config(),
            core_config.get_fleet_config(),
            task_config.get_fleet_config(),
        ]

    def get_available_instances(
        self,
        emr_client,
        releaseLabel: Optional[str] = emr_constants.releaseLabel,
        **kwargs,
    ) -> List[Any]:
        if releaseLabel is None:
            raise ValueError("releaseLabel cannot be None")
        result = []
        response = None
        while True:
            if response is None or response.get("Marker") is None:
                response = emr_client.list_supported_instance_types(
                    ReleaseLabel=releaseLabel
                )
            else:
                marker = response["Marker"]
                response = emr_client.list_supported_instance_types(
                    ReleaseLabel=releaseLabel, Marker=marker
                )
            supported_instance_types = response.get("SupportedInstanceTypes", [])
            result.extend(self.add_prefix_suffix_to_instances(supported_instance_types))
            if response.get("Marker") is None:
                break
        return result

    def get_instance_w_price(
        self, price_client, instances: Sequence[Dict], n: int = 15
    ):
        for instance in instances:
            response = price_client.get_products(
                ServiceCode="AmazonEC2",
                Filters=self.create_filters(instance),
                FormatVersion="aws_v1",
            )
            instance["Price"] = self.get_instance_price(response.get("PriceList"))
        return sorted(
            instances, key=lambda x: (x["Price"], -x["MemoryGB"], -x["VCPU"])
        )[0:n]

    def gen_instance_type_configs(self):
        instanceTypeConfigs = []
        for i in range(len(self.instanceType)):  # type: ignore
            temp = {
                "InstanceType": self.instanceType[i],  # type: ignore
                "WeightedCapacity": self.weightedCapacity[i],  # type: ignore
            }
            if self.ebsOptimized[i] is True:  # type: ignore
                ebsBlockDeviceConfigs = self._build_ebs_block_device_config(index=i)
                ebsConfiguration = {
                    "EbsBlockDeviceConfigs": ebsBlockDeviceConfigs,
                    "EbsOptimized": self.ebsOptimized[i],  # type: ignore
                }
                temp["EbsConfiguration"] = ebsConfiguration
            if self.percentageOfOnDemandPrice is not None:
                temp["BidPriceAsPercentageOfOnDemandPrice"] = (
                    self.percentageOfOnDemandPrice
                )
            elif self.bidPrice is not None:
                temp["BidPrice"] = self.bidPrice
            instanceTypeConfigs.append(temp)
        return instanceTypeConfigs

    def get_fleet_programatically(self, emrClient, priceClient):
        filter_values = [
            self.filters.get("VCPU"),  # type: ignore
            self.filters.get(  # type: ignore
                "WorkerInstanceFamilyId",
                self.filters.get("InstanceFamilyId"),  # type: ignore
            ),
            self.filters.get("MemoryGB"),  # type: ignore
            self.filters.get("WorkerPrefix", self.filters.get("Prefix")),  # type: ignore
            self.filters.get("WorkerSuffix", self.filters.get("Suffix")),  # type: ignore
        ]
        if self.filters.get("StorageGB") is None or all(  # type: ignore
            value is None for value in filter_values
        ):
            raise
        instances = self.get_available_instances(emr_client=emrClient)
        master_candidates = self.filter_instances(
            instances,
            StorageGB=(
                self.filters.get("StorageGB")  # type: ignore
                if self.filters.get("NoEBS") is True  # type: ignore
                else None
            ),
            VCPU=self.filters.get("VCPU"),  # type: ignore
            InstanceFamilyId=self.filters.get("InstanceFamilyId"),  # type: ignore
            MemoryGB=self.filters.get("MemoryGB"),  # type: ignore
            Prefix=self.filters.get("Prefix"),  # type: ignore
            Suffix=self.filters.get("Suffix"),  # type: ignore
        )
        master_candidates = self.get_instance_w_price(
            price_client=priceClient,
            instances=master_candidates,  # type: ignore
            n=5,  # type: ignore
        )
        candidates = self.filter_instances(
            instances,
            StorageGB=(
                self.filters.get("StorageGB")  # type: ignore
                if self.filters.get("NoEBS") is True  # type: ignore
                else None
            ),
            VCPU=self.filters.get("VCPU"),  # type: ignore
            InstanceFamilyId=self.filters.get(  # type: ignore
                "WorkerInstanceFamilyId",
                self.filters.get("InstanceFamilyId"),  # type: ignore
            ),
            Prefix=self.filters.get("WorkerPrefix", self.filters.get("Prefix")),  # type: ignore
            Suffix=self.filters.get("WorkerSuffix", self.filters.get("Suffix")),  # type: ignore
            MemoryGB=self.filters.get("MemoryGB"),  # type: ignore
        )
        candidates = self.get_instance_w_price(
            price_client=priceClient,
            instances=candidates,  # type: ignore
        )

        master_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.master.value,
            Market=emr_constants.Market.on_demand.value,
            Name="master",
            InstanceType=[item["Type"] for item in master_candidates],
            SizeInGB=[self.sizeInGB for _ in range(len(master_candidates))],
            WeightedCapacity=[1 for _ in range(len(master_candidates))],
            MemoryGB=[item["MemoryGB"] for item in master_candidates],
            StorageGB=[item["StorageGB"] for item in master_candidates],
            NoEBS=self.noEBS,
            VolumeType=[self.volumeType for _ in range(len(master_candidates))],
            VolumesPerInstance=[
                self.volumesPerInstance for _ in range(len(master_candidates))
            ],
            EbsOptimized=[
                (
                    item["EbsOptimizedAvailable"]
                    if item["StorageGB"] < self.filters.get("StorageGB")  # type: ignore
                    else False
                )
                for item in master_candidates
            ],
            TargetOnDemand=1,
            ReservationPreference=self.reservationPreference,
            TimeoutDuration=self.timeoutDuration,
        )

        core_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.core.value,
            Market=emr_constants.Market.on_demand.value,
            Name="core",
            InstanceType=[item["Type"] for item in candidates[0:5]],
            SizeInGB=[self.sizeInGB for _ in range(len(candidates[0:5]))],
            WeightedCapacity=[
                (
                    self.weightedCapacity
                    if self.weightedCapacity is not None
                    else item["VCPU"]
                )
                for item in candidates[0:5]
            ],
            MemoryGB=[item["MemoryGB"] for item in candidates[0:5]],
            StorageGB=[item["StorageGB"] for item in candidates[0:5]],
            NoEBS=self.noEBS,
            VolumeType=[self.volumeType for _ in range(len(candidates[0:5]))],
            VolumesPerInstance=[
                self.volumesPerInstance for _ in range(len(candidates[0:5]))
            ],
            EbsOptimized=[
                (
                    item["EbsOptimizedAvailable"]
                    if item["StorageGB"] < self.filters.get("StorageGB")  # type: ignore
                    else False
                )
                for item in candidates[0:5]
            ],
            TargetOnDemand=self.maximumCoreCapacityUnits,
            ReservationPreference=self.reservationPreference,
            TimeoutDuration=self.timeoutDuration,
        )
        task_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.task.value,
            Name="task",
            InstanceType=[item["Type"] for item in candidates],
            SizeInGB=[self.sizeInGB for _ in range(len(candidates))],
            WeightedCapacity=[
                (
                    self.weightedCapacity
                    if self.weightedCapacity is not None
                    else item["VCPU"]
                )
                for item in candidates
            ],
            MemoryGB=[item["MemoryGB"] for item in candidates],
            StorageGB=[item["StorageGB"] for item in candidates],
            NoEBS=self.noEBS if self.noEBS is not None else False,
            VolumeType=[self.volumeType for _ in range(len(candidates))],
            VolumesPerInstance=[
                self.volumesPerInstance for _ in range(len(candidates))
            ],
            EbsOptimized=[
                (
                    item["EbsOptimizedAvailable"]
                    if item["StorageGB"] < self.filters.get("StorageGB")  # type: ignore
                    else False
                )
                for item in candidates
            ],
            ReservationPreference=self.reservationPreference,
            PercentageOfOnDemandPrice=self.percentageOfOnDemandPrice,
            AllocationStrategy=(
                self.allocationStrategy
                if self.allocationStrategy is None
                else "lowest-price"
            ),
            TimeoutAction=self.timeoutAction,
            TimeoutDuration=self.timeoutDuration,
            TargetOnDemand=self.maximumOnDemandCapacityUnits,
            TargetOnSpot=self.maximumCapacityUnits
            - self.maximumCoreCapacityUnits
            - self.maximumOnDemandCapacityUnits,
        )
        get_dagster_logger().info(f"task: {task_config.get_fleet_config()}")
        return [
            master_config.get_fleet_config(),
            core_config.get_fleet_config(),
            task_config.get_fleet_config(),
        ]

    def get_group_default(self):
        master_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.master,
            Market=emr_constants.Market.on_demand,
            Name="Master Group",
            CustomAmiId=[self.customAmiId],
            InstanceType=self.instanceType,
            InstanceCount=self.masterInstanceCount,
            BidPrice=self.bidPrice,
            SizeInGB=[self.sizeInGB],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=self.ebsOptimized if self.ebsOptimized is not None else True,
        )

        core_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.core,
            Market=emr_constants.Market.on_demand,
            Name="Core Group",
            CustomAmiId=[self.customAmiId],
            InstanceType=self.workerInstanceType,
            InstanceCount=self.coreInstanceCount,
            BidPrice=self.bidPrice,
            SizeInGB=[self.sizeInGB],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=self.ebsOptimized if self.ebsOptimized is not None else True,
        )

        task_config = CloudInstanceConfig(
            InstanceRole=emr_constants.InstanceRole.task,
            Market=self.market,
            Name="Task Group",
            CustomAmiId=[self.customAmiId],
            InstanceType=self.workerInstanceType,
            InstanceCount=self.taskInstanceCount,
            BidPrice=self.bidPrice,
            SizeInGB=[self.sizeInGB],
            VolumeType=[self.volumeType],
            VolumesPerInstance=[self.volumesPerInstance],
            EbsOptimized=self.ebsOptimized if self.ebsOptimized is not None else True,
        )
        return [
            master_config.get_instance_config(),
            core_config.get_instance_config(),
            task_config.get_instance_config(),
        ]



import logging


def quiet_docling(level=logging.ERROR):
    for name in (
        "docling",
        "docling.backend.html_backend",
        "pdfminer",
        "pikepdf",
        "fitz",
        "PIL",
        "unstructured",
    ):
        lg = logging.getLogger(name)
        lg.setLevel(level)
        lg.propagate = False
        # Optional: nuke any pre-attached handlers
        if lg.handlers:
            lg.handlers.clear()



import os
from pathlib import Path

from loguru import logger


def set_cache_dir(cache_dir_base="/data/raid5/data/ascii/models/"):
    """Check if on the server and root path exists, if this is the case set the models base directory.

    > WARNING: This function MUST be executed before importing `huggingface, transformers`, ... or any other model library
    """
    data_path = Path(cache_dir_base)

    if not data_path.exists():
        logger.info("Not on server - noop not changing caching directory")
        # data_path.mkdir(parents=True, exist_ok=True)
    else:
        hf_home = data_path / "hf"
        hf_home.mkdir(parents=True, exist_ok=True)
        os.environ["HF_HOME"] = str(data_path)

        hf_datasets = hf_home / "datasets"
        hf_datasets.mkdir(parents=True, exist_ok=True)
        os.environ["HF_DATASETS_CACHE"] = str(hf_datasets)

        hf_transformers = hf_home / "models"
        hf_transformers.mkdir(parents=True, exist_ok=True)
        os.environ["TRANSFORMERS_CACHE"] = str(hf_transformers)



from enum import Enum, auto


class InstanceRole(Enum):
    master = "MASTER"
    core = "CORE"
    task = "TASK"


class TimeoutAction(Enum):
    switch = "SWITCH_TO_ON_DEMAND"
    terminate = "TERMINATE_CLUSTER"


# TODO: maybe we should make a list of valid instances because
#   a) not all of the instances work on EMR,
#   b) some are too big and we should not let the user pick a absurdly humongous instance


class Market(Enum):
    on_demand = "ON_DEMAND"
    spot = "SPOT"


class CapacityReservation(Enum):
    open = "open"
    close = "none"


class AllocationStrategy(Enum):
    priceCapacity = "price-capacity-optimized"
    capacity = "capacity-optimized"
    price = "lowest-price"
    mix = "diversified"


class InstanceFamilyId(Enum):
    balanceCur = "GENERAL_CURRENT_GEN"
    balancePrev = "GENERAL_PREVIOUS_GEN"
    acceleratedComp = "GPU_CURRENT_GEN"
    memoryOptimizedCur = "HI_MEM_CURRENT_GEN"
    memoryOptimizedPrev = "HI_MEM_PREVIOUS_GEN"
    machineLearningCur = "COMPUTE_CURRENT_GEN"
    machineLearningPrev = "COMPUTE_PREVIOUS_GEN"
    IOPScur = "STORAGE_CURRENT_GEN"
    IOPSprev = "STORAGE_PREVIOUS_GEN"


class Suffix(Enum):
    small = auto()
    medium = auto()
    large = auto()
    xlarge = auto()
    _2xlarge = auto()
    _6xlarge = auto()
    _8xlarge = auto()
    _9xlarge = auto()
    _3xlarge = auto()
    _4xlarge = auto()
    _10xlarge = auto()
    _12xlarge = auto()
    _16xlarge = auto()
    _18xlarge = auto()
    _24xlarge = auto()
    _32xlarge = auto()
    _48xlarge = auto()

    @classmethod
    def suffix_order(cls):
        return list(cls)

    @classmethod
    def index_of(cls, suffix_str):
        # Adjust the suffix_str to match the enum naming convention that it cannot start with number
        # If the suffix_str starts with a digit, prepend an underscore
        if suffix_str[0].isdigit():
            suffix_str = f"_{suffix_str}"
        try:
            return cls.suffix_order().index(cls[suffix_str])  # pyrefly: ignore
        except KeyError:
            return -1


class VolumeType(Enum):
    generalPurpose3 = "gp3"
    generalPurpose2 = "gp2"
    IOPS1 = "io1"
    throughputOptimizedHDD = "st1"
    coldHDD = "sc1"
    magnetic = "standard"


class Subnets(Enum):
    usEast1b = "subnet-02035452bafe54090"  # main
    usEast1e = "subnet-04be55a5e7990e7fe"
    usEast1d = "subnet-0deb46a6dcbcb90be"
    usEast1a = "subnet-01c57c11d6c92a586"
    usEast1c = "subnet-087f4559eec6caddb"
    usEast1f = "subnet-01e0cb6f08abb93ee"


pipeline_bucket = "ascii-supply-chain-research-pipeline"
mock = False
sizeInGB = 300
volumeType = VolumeType.generalPurpose2
volumesPerInstance = 1
ebsOptimized = True
percentageOfOnDemandPrice = 70.0
allocationStrategy = AllocationStrategy.priceCapacity
timeoutDuration = 10
weightedCapacity = (
    4  # this should be a number greater than 1 and should match the instance vcore
)
releaseLabel = "emr-7.10.0"



import json
import os
from typing import Any, Mapping, Optional, Tuple

from dagster import AssetKey, MetadataValue, OpExecutionContext, Output
from dagster_dbt import (  # pyrefly: ignore
    DagsterDbtTranslator,
    DagsterDbtTranslatorSettings,
    DbtCliInvocation,
    DbtCliResource,
    default_metadata_from_dbt_resource_props,
)

# all based on https://github.com/dagster-io/hooli-data-eng-pipelines/blob/master/hooli_data_eng/assets/dbt_assets.py


def build_DbtTranslator(warehouse_name: str, dbt_target_schema: str):  # noqa: C901
    class AsciiDbtTranslator(DagsterDbtTranslator):
        _partitioning_default_column_name = None
        _partitioning_overrides: Mapping[str, str] = {}

        def __init__(
            self,
            settings: Optional[DagsterDbtTranslatorSettings] = None,
            partitioning_default_column_name: str = "day_dt",
            partitioning_overrides: Mapping[str, str] = {},
        ):
            super().__init__(settings=settings)
            self._partitioning_default_column_name = (
                partitioning_default_column_name  # pyrefly: ignore
            )
            self._partitioning_overrides = partitioning_overrides

        # for details see:
        # https://github.com/dagster-io/hooli-data-eng-pipelines/blob/master/hooli_data_eng/assets/dbt_assets.py
        def get_asset_key(self, dbt_resource_props: Mapping[str, Any]) -> AssetKey:  # type: ignore
            # take key from meta if provided otherwise build unique id
            if (
                dbt_resource_props.get("meta", {})
                .get("dagster", {})
                .get("asset_key", [])
                != []
            ):
                return super().get_asset_key(dbt_resource_props)
            else:
                model_name = dbt_resource_props["name"]
                if dbt_resource_props.get("version", None) is not None:
                    model_name = model_name + "_v" + str(dbt_resource_props["version"])
                return AssetKey(model_name).with_prefix(
                    [dbt_resource_props["database"], dbt_resource_props["schema"]]
                )

        def get_group_name(self, dbt_resource_props: Mapping[str, Any]):
            "form DBT folders as asset groups. Potentially rethink in the future"
            group_name = super().get_group_name(dbt_resource_props)
            if group_name is None:
                node_path = dbt_resource_props["path"]
                if node_path == "":
                    return "upstream"
                prefix = node_path.split(os.sep)[0] if "/" in node_path else None
                if node_path.startswith("models/source_"):
                    prefix = "RAW_DATA"
                return prefix
            else:
                return group_name

        @classmethod
        def get_metadata(
            cls, dbt_resource_props: Mapping[str, Any]
        ) -> Mapping[str, Any]:
            metadata = {"partition_expr": cls._partitioning_default_column_name}

            # for any deviating partitoining name a) either load the translator with different settings
            #  or b) modify it to accept an dictionary of overrides
            for k, v in cls._partitioning_overrides.items():
                if dbt_resource_props["name"] == k:
                    metadata = {"partition_expr": v}

            default_metadata = default_metadata_from_dbt_resource_props(
                dbt_resource_props
            )

            return {**default_metadata, **metadata}  # pyrefly: ignore

    return AsciiDbtTranslator


def generate_additional_metadata_for_output(
    output: Output,
    manifest: dict,
    results_by_asset_key: dict,
    dbt_resource_props_by_asset_key: dict,
    dagster_dbt_translator: DagsterDbtTranslator,
) -> Mapping[str, Any]:
    # Get the unique id of the dbt node from the output
    value = output.metadata["unique_id"].value
    if not isinstance(value, str):
        raise TypeError(
            f"Expected unique_id to be a str, got {type(value)} instead. unique_id: {value}"
        )
    dbt_unique_id: str = value

    # Get the asset key from the unique id
    asset_key = dagster_dbt_translator.get_asset_key(manifest["nodes"][dbt_unique_id])

    # Retrieve the results using the asset key, and get the rows_affected.
    result = results_by_asset_key[asset_key]
    rows_affected: Optional[int] = result["adapter_response"].get("rows_affected")
    rows_affected_metadata = {"rows_affected": rows_affected} if rows_affected else {}

    # Retrieve the dbt resource props using the asset key, and get the compiled code and owner.
    node = dbt_resource_props_by_asset_key[asset_key]
    compiled_sql: Optional[str] = node.get("compiled_code")
    compiled_sql_metadata = (
        {"compiled_sql": MetadataValue.md(compiled_sql)} if compiled_sql else {}
    )

    meta_owner: Optional[str] = node.get("meta").get("owner")
    meta_owner_metadata = {"owner": meta_owner} if meta_owner else {}

    return {
        **rows_affected_metadata,
        **compiled_sql_metadata,
        **meta_owner_metadata,
    }


def prepare_dbt_files(
    dbt_cli_task: DbtCliInvocation, dagster_dbt_translator: DagsterDbtTranslator
) -> Tuple[dict, dict]:
    run_results = dbt_cli_task.get_artifact("run_results.json")
    executed_manifest = dbt_cli_task.get_artifact("manifest.json")

    results_by_asset_key = {
        dagster_dbt_translator.get_asset_key(
            executed_manifest["nodes"][result["unique_id"]]
        ): result
        for result in run_results["results"]
    }
    dbt_resource_props_by_asset_key = {
        dagster_dbt_translator.get_asset_key(node): node
        for node in executed_manifest["nodes"].values()
    }

    return results_by_asset_key, dbt_resource_props_by_asset_key


def process_dbt_assets(
    context: OpExecutionContext,
    dbt2: DbtCliResource,
    dagster_dbt_translator2: DagsterDbtTranslator,
    dbt_mode: str = "build",  # choose run or build
):
    if context.has_partition_key:
        # map partition key range to dbt vars
        (
            first_partition,
            last_partition,
        ) = context.asset_partitions_time_window_for_output(
            list(context.selected_output_names)[0]
        )
        # This assumes a single backfill is possible
        # in case of dependent operations i.e. SCD2 you must disable the single run option
        dbt_vars = {"min_date": str(first_partition), "max_date": str(last_partition)}
        dbt_args = [dbt_mode, "--vars", json.dumps(dbt_vars)]
        dbt_cli_task = dbt2.cli(dbt_args, context=context, raise_on_error=False)
    else:
        dbt_cli_task = dbt2.cli([dbt_mode], context=context, raise_on_error=False)

    dbt_events = list(dbt_cli_task.stream())

    results_by_asset_key, dbt_resource_props_by_asset_key = prepare_dbt_files(
        dbt_cli_task, dagster_dbt_translator2
    )
    for dagster_event in dbt_events:
        if isinstance(dagster_event, Output):
            metadata = generate_additional_metadata_for_output(
                output=dagster_event,
                manifest=dbt_cli_task.get_artifact("manifest.json"),
                results_by_asset_key=results_by_asset_key,
                dbt_resource_props_by_asset_key=dbt_resource_props_by_asset_key,
                dagster_dbt_translator=dagster_dbt_translator2,
            )
            context.add_output_metadata(
                metadata=metadata,
                output_name=dagster_event.output_name,
            )
        yield dagster_event



from contextlib import contextmanager
from typing import Generator

import dagster as dg
from duckdb import DuckDBPyConnection  # pyrefly: ignore


class DuckDBConnectionProvider(dg.ConfigurableResource):
    """
    A protocol for Dagster resources that can provide a native DuckDB connection.

    This abstraction allows ops to depend on a generic connection provider,
    making them substitutable between different backends like a local file
    or a remote DuckLake instance.
    """

    @contextmanager
    def duckdb_connect(self) -> Generator[DuckDBPyConnection, None, None]:
        """
        A context manager that yields a pre-configured DuckDB connection.
        The connection is automatically closed upon exiting the context.
        """
        ...



# pyright: reportArgumentType=false, reportCallIssue=false, reportAssignmentType=false
from typing import Optional

import boto3
from dagster import ConfigurableResource
from obstore.auth.boto3 import Boto3CredentialProvider  # pyrefly: ignore
from obstore.store import S3Store  # pyrefly: ignore
from pydantic import Field


class ObstoreS3Resource(ConfigurableResource):
    """Dagster integration for fast rust based s3 client."""

    region_name: Optional[str] = Field(
        default=None, description="Specifies a custom region for the S3 session."
    )
    endpoint_url: Optional[str] = Field(
        default=None, description="Specifies a custom endpoint for the S3 session."
    )
    profile_name: Optional[str] = Field(
        default=None, description="Specifies a profile to connect that session."
    )
    aws_access_key_id: Optional[str] = Field(
        default=None,
        description="AWS access key ID to use when creating the boto3 session.",
    )
    aws_secret_access_key: Optional[str] = Field(
        default=None,
        description="AWS secret access key to use when creating the boto3 session.",
    )
    allow_http: bool = Field(
        default=False,
        description="Whether to allow http connections. By default, https is used.",
    )
    allow_invalid_certificates: bool = Field(
        default=False,
        description="Whether to allow invalid certificates. By default, valid certs are required.",
    )

    def get_client(
        self,
        bucket: str,
        timeout: str = "60s",
        retry_config=None,
    ) -> S3Store:
        """Creates an S3 object store."""
        config = {
            k: v
            for k, v in {
                "aws_access_key_id": self.aws_access_key_id,
                "aws_secret_access_key": self.aws_secret_access_key,
                "aws_endpoint": self.endpoint_url,
                "aws_region": self.region_name,
                "aws_bucket": bucket,
            }.items()
            if v is not None
        }
        client_options = {
            "timeout": timeout,
            "allow_http": self.allow_http,
            "allow_invalid_certificates": self.allow_invalid_certificates,
        }
        if self.profile_name is not None:
            client_session = boto3.session.Session(profile_name=self.profile_name)
            credential_provider = Boto3CredentialProvider(client_session)
            return S3Store(
                credential_provider=credential_provider,
                config=config,
                retry_config=retry_config,
                client_options=client_options,
            )
        else:
            return S3Store(
                config=config,
                retry_config=retry_config,
                client_options=client_options,
            )
